# Copyright (c) 2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from tempfile import NamedTemporaryFile
import json, urllib
from operator import itemgetter

from warcio.archiveiterator import ArchiveIterator
from yarl import URL
from multidict import CIMultiDict
from hypothesis import given, reproduce_failure
import hypothesis.strategies as st
import pytest

from .warc import WarcHandler
from .logger import Logger, WarcHandlerConsumer
from .controller import ControllerStart
from .behavior import Script, ScreenshotEvent, DomSnapshotEvent
from .browser import RequestResponsePair, Base64Body, UnicodeBody
from .test_browser import requestResponsePair, urls

def test_log ():
    logger = Logger ()

    with NamedTemporaryFile() as fd:
        with WarcHandler (fd, logger) as handler:
            warclogger = WarcHandlerConsumer (handler)
            logger.connect (warclogger)
            golden = []

            assert handler.log.tell () == 0
            golden.append (logger.info (foo=1, bar='baz', encoding='äöü⇔ΓΨ'))
            assert handler.log.tell () != 0

            handler.maxLogSize = 0
            golden.append (logger.info (bar=1, baz='baz'))
            # should flush the log
            assert handler.log.tell () == 0

        fd.seek (0)
        for it in ArchiveIterator (fd):
            headers = it.rec_headers
            assert headers['warc-type'] == 'metadata'
            assert 'warc-target-uri' not in headers
            assert headers['x-crocoite-type'] == 'log'
            assert headers['content-type'] == f'application/json; encoding={handler.logEncoding}'

            while True:
                l = it.raw_stream.readline ()
                if not l:
                    break
                data = json.loads (l.strip ())
                assert data == golden.pop (0)

def jsonObject ():
    """ JSON-encodable objects """
    return st.dictionaries (st.text (), st.one_of (st.integers (), st.text ()))

def viewport ():
    return st.builds (lambda x, y: f'{x}x{y}', st.integers (), st.integers ())

def event ():
    return st.one_of (
            st.builds (ControllerStart, jsonObject ()),
            st.builds (Script.fromStr, st.text (), st.one_of(st.none (), st.text ())),
            st.builds (ScreenshotEvent, urls (), st.integers (), st.binary ()),
            st.builds (DomSnapshotEvent, urls (), st.builds (lambda x: x.encode ('utf-8'), st.text ()), viewport()),
            requestResponsePair (),
            )

@pytest.mark.asyncio
@given (st.lists (event ()))
async def test_push (golden):
    def checkWarcinfoId (headers):
        if lastWarcinfoRecordid is not None:
            assert headers['WARC-Warcinfo-ID'] == lastWarcinfoRecordid

    lastWarcinfoRecordid = None

    # null logger
    logger = Logger ()
    with open('/tmp/test.warc.gz', 'w+b') as fd:
        with WarcHandler (fd, logger) as handler:
            for g in golden:
                await handler.push (g)

        fd.seek (0)
        it = iter (ArchiveIterator (fd))
        for g in golden:
            if isinstance (g, ControllerStart):
                rec = next (it)

                headers = rec.rec_headers
                assert headers['warc-type'] == 'warcinfo'
                assert 'warc-target-uri' not in headers
                assert 'x-crocoite-type' not in headers

                data = json.load (rec.raw_stream)
                assert data == g.payload

                lastWarcinfoRecordid = headers['warc-record-id']
                assert lastWarcinfoRecordid
            elif isinstance (g, Script):
                rec = next (it)

                headers = rec.rec_headers
                assert headers['warc-type'] == 'resource'
                assert headers['content-type'] == 'application/javascript; charset=utf-8'
                assert headers['x-crocoite-type'] == 'script'
                checkWarcinfoId (headers)
                if g.path:
                    assert URL (headers['warc-target-uri']) == URL ('file://' + g.abspath)
                else:
                    assert 'warc-target-uri' not in headers

                data = rec.raw_stream.read ().decode ('utf-8')
                assert data == g.data
            elif isinstance (g, ScreenshotEvent):
                # XXX: check refers-to header
                rec = next (it)

                headers = rec.rec_headers
                assert headers['warc-type'] == 'conversion'
                assert headers['x-crocoite-type'] == 'screenshot'
                checkWarcinfoId (headers)
                assert URL (headers['warc-target-uri']) == g.url, (headers['warc-target-uri'], g.url)
                assert headers['warc-refers-to'] is None
                assert int (headers['X-Crocoite-Screenshot-Y-Offset']) == g.yoff

                assert rec.raw_stream.read () == g.data
            elif isinstance (g, DomSnapshotEvent):
                rec = next (it)

                headers = rec.rec_headers
                assert headers['warc-type'] == 'conversion'
                assert headers['x-crocoite-type'] == 'dom-snapshot'
                checkWarcinfoId (headers)
                assert URL (headers['warc-target-uri']) == g.url
                assert headers['warc-refers-to'] is None

                assert rec.raw_stream.read () == g.document
            elif isinstance (g, RequestResponsePair):
                rec = next (it)

                # request
                headers = rec.rec_headers
                assert headers['warc-type'] == 'request'
                assert 'x-crocoite-type' not in headers
                checkWarcinfoId (headers)
                assert URL (headers['warc-target-uri']) == g.url
                assert headers['x-chrome-request-id'] == g.id
                
                assert CIMultiDict (rec.http_headers.headers) == g.request.headers
                if g.request.hasPostData:
                    if g.request.body is not None:
                        assert rec.raw_stream.read () == g.request.body
                    else:
                        # body fetch failed
                        assert headers['warc-truncated'] == 'unspecified'
                        assert not rec.raw_stream.read ()
                else:
                    assert not rec.raw_stream.read ()

                # response
                if g.response:
                    rec = next (it)
                    headers = rec.rec_headers
                    httpheaders = rec.http_headers
                    assert headers['warc-type'] == 'response'
                    checkWarcinfoId (headers)
                    assert URL (headers['warc-target-uri']) == g.url
                    assert headers['x-chrome-request-id'] == g.id
                    assert 'x-crocoite-type' not in headers

                    # these are checked separately
                    filteredHeaders = CIMultiDict (httpheaders.headers)
                    for b in {'content-type', 'content-length'}:
                        if b in g.response.headers:
                            g.response.headers.popall (b)
                        if b in filteredHeaders:
                            filteredHeaders.popall (b)
                    assert filteredHeaders == g.response.headers

                    expectedContentType = g.response.mimeType
                    if expectedContentType is not None:
                        assert httpheaders['content-type'].startswith (expectedContentType)

                    if g.response.body is not None:
                        assert rec.raw_stream.read () == g.response.body
                        assert httpheaders['content-length'] == str (len (g.response.body))
                        # body is never truncated if it exists
                        assert headers['warc-truncated'] is None

                        # unencoded strings are converted to utf8
                        if isinstance (g.response.body, UnicodeBody) and httpheaders['content-type'] is not None:
                            assert httpheaders['content-type'].endswith ('; charset=utf-8')
                    else:
                        # body fetch failed
                        assert headers['warc-truncated'] == 'unspecified'
                        assert not rec.raw_stream.read ()
                        # content-length header should be kept intact
            else:
                assert False, f"invalid golden type {type(g)}" # pragma: no cover

        # no further records
        with pytest.raises (StopIteration):
            next (it)




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
HTML helper
"""

from html5lib.treewalkers.base import TreeWalker
from html5lib.filters.base import Filter
from html5lib import constants

# HTML void tags, see https://html.spec.whatwg.org/multipage/syntax.html#void-elements
voidTags = {'area',
        'base',
        'br',
        'col',
        'embed',
        'hr',
        'img',
        'input',
        'link',
        'meta',
        'param',
        'source',
        'track',
        'wbr'}

# source: https://developer.mozilla.org/en-US/docs/Web/HTML/Global_attributes
eventAttributes = {'onabort',
        'onautocomplete',
        'onautocompleteerror',
        'onblur',
        'oncancel',
        'oncanplay',
        'oncanplaythrough',
        'onchange',
        'onclick',
        'onclose',
        'oncontextmenu',
        'oncuechange',
        'ondblclick',
        'ondrag',
        'ondragend',
        'ondragenter',
        'ondragexit',
        'ondragleave',
        'ondragover',
        'ondragstart',
        'ondrop',
        'ondurationchange',
        'onemptied',
        'onended',
        'onerror',
        'onfocus',
        'oninput',
        'oninvalid',
        'onkeydown',
        'onkeypress',
        'onkeyup',
        'onload',
        'onloadeddata',
        'onloadedmetadata',
        'onloadstart',
        'onmousedown',
        'onmouseenter',
        'onmouseleave',
        'onmousemove',
        'onmouseout',
        'onmouseover',
        'onmouseup',
        'onmousewheel',
        'onpause',
        'onplay',
        'onplaying',
        'onprogress',
        'onratechange',
        'onreset',
        'onresize',
        'onscroll',
        'onseeked',
        'onseeking',
        'onselect',
        'onshow',
        'onsort',
        'onstalled',
        'onsubmit',
        'onsuspend',
        'ontimeupdate',
        'ontoggle',
        'onvolumechange',
        'onwaiting'}

default_namespace = constants.namespaces["html"]

class ChromeTreeWalker (TreeWalker):
    """
    Recursive html5lib TreeWalker for Google Chrome method DOM.getDocument
    """

    def recurse (self, node):
        name = node['nodeName']
        if name.startswith ('#'):
            if name == '#text':
                yield from self.text (node['nodeValue'])
            elif name == '#comment':
                yield self.comment (node['nodeValue'])
            elif name == '#document':
                for child in node.get ('children', []):
                    yield from self.recurse (child)
            elif name == '#cdata-section':
                # html5lib cannot generate cdata, so we’re faking it by using
                # an empty tag
                yield from self.emptyTag (default_namespace,
                        '![CDATA[' + node['nodeValue'] + ']]', {})
            else:
                assert False, (name, node)
        else:
            attributes = node.get ('attributes', [])
            convertedAttr = {}
            for i in range (0, len (attributes), 2):
                convertedAttr[(default_namespace, attributes[i])] = attributes[i+1]

            children = node.get ('children', [])
            if name.lower() in voidTags and not children:
                yield from self.emptyTag (default_namespace, name, convertedAttr)
            else:
                yield self.startTag (default_namespace, name, convertedAttr)
                for child in node.get ('children', []):
                    yield from self.recurse (child)
                yield self.endTag ('', name)

    def __iter__ (self):
        assert self.tree['nodeName'] == '#document'
        return self.recurse (self.tree)

    def split (self):
        """
        Split response returned by DOM.getDocument(pierce=True) into independent documents
        """
        def recurse (node):
            contentDocument = node.get ('contentDocument')
            if contentDocument:
                assert contentDocument['nodeName'] == '#document'
                yield contentDocument
                yield from recurse (contentDocument)

            for child in node.get ('children', []):
                yield from recurse (child)

        if self.tree['nodeName'] == '#document':
            yield self.tree
        yield from recurse (self.tree)

class StripTagFilter (Filter):
    """
    Remove arbitrary tags
    """

    def __init__ (self, source, tags):
        Filter.__init__ (self, source)
        self.tags = set (map (str.lower, tags))

    def __iter__(self):
        delete = 0
        for token in Filter.__iter__(self):
            tokenType = token['type']
            if tokenType in {'StartTag', 'EmptyTag'}:
                if delete > 0 or token['name'].lower () in self.tags:
                    delete += 1
            if delete == 0:
                yield token
            if tokenType == 'EndTag' and delete > 0:
                delete -= 1

class StripAttributeFilter (Filter):
    """
    Remove arbitrary HTML attributes
    """

    def __init__ (self, source, attributes):
        Filter.__init__ (self, source)
        self.attributes = set (map (str.lower, attributes))

    def __iter__(self):
        for token in Filter.__iter__(self):
            data = token.get ('data')
            if data and token['type'] in {'StartTag', 'EmptyTag'}:
                newdata = {}
                for (namespace, k), v in data.items ():
                    if k.lower () not in self.attributes:
                        newdata[(namespace, k)] = v
                token['data'] = newdata
            yield token




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio
import pytest

from aiohttp import web
import websockets

from .devtools import Browser, Tab, MethodNotFound, Crashed, \
        InvalidParameter, Process, Passthrough

@pytest.fixture
async def browser ():
    async with Process () as url:
        yield Browser (url)

@pytest.fixture
async def tab (browser):
    async with browser as tab:
        yield tab
        # make sure there are no transactions left over (i.e. no unawaited requests)
        assert not tab.transactions

docBody = "<html><body><p>Hello, world</p></body></html>"
async def hello(request):
    return web.Response(text=docBody, content_type='text/html')

@pytest.fixture
async def server ():
    """ Simple HTTP server for testing notifications """
    app = web.Application()
    app.add_routes([web.get('/', hello)])
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, 'localhost', 8080)
    await site.start()
    yield app
    await runner.cleanup ()

@pytest.mark.asyncio
async def test_tab_create (tab):
    """ Creating tabs works """
    assert isinstance (tab, Tab)
    version = await tab.Browser.getVersion ()
    assert 'protocolVersion' in version
    assert tab.pending == 0

@pytest.mark.asyncio
async def test_tab_close (browser):
    """ Tabs are closed after using them """
    async with browser as tab:
        tid = tab.id
    # give the browser some time to close the tab
    await asyncio.sleep (0.5)
    tabs = [t['id'] async for t in browser]
    assert tid not in tabs

@pytest.mark.asyncio
async def test_tab_notify_enable_disable (tab):
    """ Make sure enabling/disabling notifications works for all known
    namespaces """
    for name in ('Debugger', 'DOM', 'Log', 'Network', 'Page', 'Performance',
            'Profiler', 'Runtime', 'Security'):
        f = getattr (tab, name)
        await f.enable ()
        await f.disable ()

@pytest.mark.asyncio
async def test_tab_unknown_method (tab):
    with pytest.raises (MethodNotFound):
        await tab.Nonexistent.foobar ()

@pytest.mark.asyncio
async def test_tab_invalid_argument (tab):
    # should be string
    with pytest.raises (InvalidParameter):
        await tab.Page.captureScreenshot (format=123)

    with pytest.raises (InvalidParameter):
        await tab.Page.captureScreenshot (format=[123])

    with pytest.raises (InvalidParameter):
        await tab.Page.captureScreenshot (format={123: '456'})

@pytest.mark.asyncio
async def test_tab_crash (tab):
    with pytest.raises (Crashed):
        await tab.Page.crash ()

    # caling anything else now should fail as well
    with pytest.raises (Crashed):
        await tab.Browser.getVersion ()

@pytest.mark.asyncio
async def test_load (tab, server):
    await tab.Network.enable ()
    await tab.Page.navigate (url='http://localhost:8080')

    method, req = await tab.get ()
    assert method == tab.Network.requestWillBeSent

    method, resp = await tab.get ()
    assert method == tab.Network.responseReceived
    assert resp['requestId'] == req['requestId']

    method, dataRecv = await tab.get ()
    assert method == tab.Network.dataReceived
    assert dataRecv['dataLength'] == len (docBody)
    assert dataRecv['requestId'] == req['requestId']

    method, finish = await tab.get ()
    assert method == tab.Network.loadingFinished
    assert finish['requestId'] == req['requestId']

    body = await tab.Network.getResponseBody (requestId=req['requestId'])
    assert body['body'] == docBody

    await tab.Network.disable ()
    assert tab.pending == 0

@pytest.mark.asyncio
async def test_recv_failure(browser):
    """ Inject failure into receiver process and crash it """
    async with browser as tab:
        await tab.ws.close ()
        with pytest.raises (Crashed):
            await tab.Browser.getVersion ()

    async with browser as tab:
        await tab.ws.close ()
        with pytest.raises (Crashed):
            await tab.get ()

    async with browser as tab:
        handle = asyncio.ensure_future (tab.get ())
        await tab.ws.close ()
        with pytest.raises (Crashed):
            await handle

@pytest.mark.asyncio
async def test_tab_function (tab):
    assert tab.Network.enable.name == 'Network.enable'
    assert tab.Network.disable == tab.Network.disable
    assert tab.Network.enable != tab.Network.disable
    assert tab.Network != tab.Network.enable
    assert callable (tab.Network.enable)
    assert not callable (tab.Network.enable.name)
    assert 'Network.enable' in repr (tab.Network.enable)

@pytest.mark.asyncio
async def test_tab_function_hash (tab):
    d = {tab.Network.enable: 1, tab.Network.disable: 2, tab.Page: 3,
            tab.Page.enable: 4}
    assert len (d) == 4

@pytest.mark.asyncio
async def test_ws_ping(tab):
    """
    Chrome does not like websocket pings and closes the connection if it
    receives one. Not sure why.
    """
    with pytest.raises (Crashed):
        await tab.ws.ping ()
        await tab.Browser.getVersion ()

@pytest.mark.asyncio
async def test_passthrough ():
    """ Null service returns the url as is """

    url = 'http://localhost:12345'
    async with Passthrough (url) as u:
        assert str (u) == url




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Command line interface
"""

import argparse, sys, signal, asyncio, os, json
from traceback import TracebackException
from enum import IntEnum
from yarl import URL
try:
    import manhole
    manhole.install (patch_fork=False, oneshot_on='USR1')
except ModuleNotFoundError:
    pass

from . import behavior, browser
from .controller import SinglePageController, \
        ControllerSettings, StatsHandler, LogHandler, \
        RecursiveController, DepthLimit, PrefixLimit
from .devtools import Passthrough, Process
from .warc import WarcHandler
from .logger import Logger, JsonPrintConsumer, DatetimeConsumer, \
        WarcHandlerConsumer, Level
from .devtools import Crashed

def absurl (s):
    """ argparse: Absolute URL """
    u = URL (s)
    if u.is_absolute ():
        return u
    raise argparse.ArgumentTypeError ('Must be absolute')

class SingleExitStatus(IntEnum):
    """ Exit status for single-shot command line """
    Ok = 0
    Fail = 1
    BrowserCrash = 2
    Navigate = 3

def single ():
    parser = argparse.ArgumentParser(description='crocoite helper tools to fetch individual pages.')
    parser.add_argument('--browser', help='DevTools URL', type=absurl, metavar='URL')
    parser.add_argument('--timeout', default=1*60*60, type=int, help='Maximum time for archival', metavar='SEC')
    parser.add_argument('--idle-timeout', default=30, type=int, help='Maximum idle seconds (i.e. no requests)', dest='idleTimeout', metavar='SEC')
    parser.add_argument('--behavior', help='Enable behavior script',
            dest='enabledBehaviorNames',
            default=list (behavior.availableMap.keys ()),
            choices=list (behavior.availableMap.keys ()),
            metavar='NAME', nargs='*')
    parser.add_argument('--warcinfo', help='Add extra information to warcinfo record',
            metavar='JSON', type=json.loads)
    parser.add_argument('-k', '--insecure',
            action='store_true',
            help='Disable certificate validation')
    parser.add_argument('url', help='Website URL', type=absurl, metavar='URL')
    parser.add_argument('output', help='WARC filename', metavar='FILE')

    args = parser.parse_args ()

    logger = Logger (consumer=[DatetimeConsumer (), JsonPrintConsumer ()])

    ret = SingleExitStatus.Fail
    service = Process ()
    if args.browser:
        service = Passthrough (args.browser)
    settings = ControllerSettings (
            idleTimeout=args.idleTimeout,
            timeout=args.timeout,
            insecure=args.insecure,
            )
    with open (args.output, 'wb') as fd, WarcHandler (fd, logger) as warcHandler:
        logger.connect (WarcHandlerConsumer (warcHandler))
        handler = [StatsHandler (), LogHandler (logger), warcHandler]
        b = list (map (lambda x: behavior.availableMap[x], args.enabledBehaviorNames))
        controller = SinglePageController (url=args.url, settings=settings,
                service=service, handler=handler, behavior=b, logger=logger,
                warcinfo=args.warcinfo)
        try:
            loop = asyncio.get_event_loop()
            run = asyncio.ensure_future (controller.run ())
            stop = lambda signum: run.cancel ()
            loop.add_signal_handler (signal.SIGINT, stop, signal.SIGINT)
            loop.add_signal_handler (signal.SIGTERM, stop, signal.SIGTERM)
            loop.run_until_complete(run)
            loop.close()
            ret = SingleExitStatus.Ok
        except Crashed:
            ret = SingleExitStatus.BrowserCrash
        except asyncio.CancelledError:
            # don’t log this one
            pass
        except browser.NavigateError:
            ret = SingleExitStatus.Navigate
        except Exception as e:
            ret = SingleExitStatus.Fail
            logger.error ('cli exception',
                    uuid='7fd69858-ecaa-4225-b213-8ab880aa3cc5',
                    traceback=list (TracebackException.from_exception (e).format ()))
        finally:
            r = handler[0].stats
            logger.info ('stats', context='cli', uuid='24d92d16-770e-4088-b769-4020e127a7ff', **r)
            logger.info ('exit', context='cli', uuid='9b1bd603-f7cd-4745-895a-5b894a5166f2', status=ret)

    return ret

def parsePolicy (recursive, url):
    if recursive is None:
        return DepthLimit (0)
    elif recursive.isdigit ():
        return DepthLimit (int (recursive))
    elif recursive == 'prefix':
        return PrefixLimit (url)
    raise argparse.ArgumentTypeError ('Unsupported recursion mode')

def recursive ():
    logger = Logger (consumer=[DatetimeConsumer (), JsonPrintConsumer ()])

    parser = argparse.ArgumentParser(description='Save website to WARC using Google Chrome.')
    parser.add_argument('-j', '--concurrency',
            help='Run at most N jobs concurrently', metavar='N', default=1,
            type=int)
    parser.add_argument('-r', '--recursion', help='Recursion policy',
            metavar='POLICY')
    parser.add_argument('--tempdir', help='Directory for temporary files',
            metavar='DIR')
    parser.add_argument('url', help='Seed URL', type=absurl, metavar='URL')
    parser.add_argument('output',
            help='Output file, supports templates {host}, {date} and {seqnum}',
            metavar='FILE')
    parser.add_argument('command',
            help='Fetch command, supports templates {url} and {dest}',
            metavar='CMD', nargs='*',
            default=['crocoite-single', '{url}', '{dest}'])

    args = parser.parse_args ()
    try:
        policy = parsePolicy (args.recursion, args.url)
    except argparse.ArgumentTypeError as e:
        parser.error (str (e))

    try:
        controller = RecursiveController (url=args.url, output=args.output,
                command=args.command, logger=logger, policy=policy,
                tempdir=args.tempdir, concurrency=args.concurrency)
    except ValueError as e:
        parser.error (str (e))

    run = asyncio.ensure_future (controller.run ())
    loop = asyncio.get_event_loop()
    stop = lambda signum: run.cancel ()
    loop.add_signal_handler (signal.SIGINT, stop, signal.SIGINT)
    loop.add_signal_handler (signal.SIGTERM, stop, signal.SIGTERM)
    loop.run_until_complete(run)
    loop.close()

    return 0

def irc ():
    import json, re
    from .irc import Chromebot

    logger = Logger (consumer=[DatetimeConsumer (), JsonPrintConsumer ()])

    parser = argparse.ArgumentParser(description='IRC bot.')
    parser.add_argument('--config', '-c', help='Config file location', metavar='PATH', default='chromebot.json')

    args = parser.parse_args ()

    with open (args.config) as fd:
        config = json.load (fd)
    s = config['irc']
    blacklist = dict (map (lambda x: (re.compile (x[0], re.I), x[1]), config['blacklist'].items ()))

    loop = asyncio.get_event_loop()
    bot = Chromebot (
            host=s['host'],
            port=s['port'],
            ssl=s['ssl'],
            nick=s['nick'],
            channels=s['channels'],
            tempdir=config['tempdir'],
            destdir=config['destdir'],
            processLimit=config['process_limit'],
            logger=logger,
            blacklist=blacklist,
            needVoice=config['need_voice'],
            loop=loop)
    stop = lambda signum: bot.cancel ()
    loop.add_signal_handler (signal.SIGINT, stop, signal.SIGINT)
    loop.add_signal_handler (signal.SIGTERM, stop, signal.SIGTERM)
    loop.run_until_complete(bot.run ())

def dashboard ():
    from .irc import Dashboard

    loop = asyncio.get_event_loop()
    d = Dashboard (sys.stdin, loop)
    loop.run_until_complete(d.run ())
    loop.run_forever()




# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio

from yarl import URL
from aiohttp import web

import pytest

from .logger import Logger
from .controller import ControllerSettings, SinglePageController, SetEntry, \
        IdleStateTracker
from .browser import PageIdle
from .devtools import Process
from .test_browser import loader

@pytest.mark.asyncio
async def test_controller_timeout ():
    """ Make sure the controller terminates, even if the site keeps reloading/fetching stuff """

    async def f (req):
        return web.Response (body="""<html>
<body>
<p>hello</p>
<script>
window.setTimeout (function () { window.location = '/' }, 250);
window.setInterval (function () { fetch('/').then (function (e) { console.log (e) }) }, 150);
</script>
</body>
</html>""", status=200, content_type='text/html', charset='utf-8')

    url = URL.build (scheme='http', host='localhost', port=8080)
    app = web.Application ()
    app.router.add_route ('GET', '/', f)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, url.host, url.port)
    await site.start()

    loop = asyncio.get_event_loop ()
    try:
        logger = Logger ()
        settings = ControllerSettings (idleTimeout=1, timeout=5)
        controller = SinglePageController (url=url, logger=logger,
                service=Process (), behavior=[], settings=settings)
        # give the controller a little more time to finish, since there are
        # hard-coded asyncio.sleep calls in there right now.
        # XXX fix this
        before = loop.time ()
        await asyncio.wait_for (controller.run (), timeout=settings.timeout*2)
        after = loop.time ()
        assert after-before >= settings.timeout, (settings.timeout*2, after-before)
    finally:
        # give the browser some time to close before interrupting the
        # connection by destroying the HTTP server
        await asyncio.sleep (1)
        await runner.cleanup ()

@pytest.mark.asyncio
async def test_controller_idle_timeout ():
    """ Make sure the controller terminates, even if the site keeps reloading/fetching stuff """

    async def f (req):
        return web.Response (body="""<html>
<body>
<p>hello</p>
<script>
window.setInterval (function () { fetch('/').then (function (e) { console.log (e) }) }, 2000);
</script>
</body>
</html>""", status=200, content_type='text/html', charset='utf-8')

    url = URL.build (scheme='http', host='localhost', port=8080)
    app = web.Application ()
    app.router.add_route ('GET', '/', f)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, url.host, url.port)
    await site.start()

    loop = asyncio.get_event_loop ()
    try:
        logger = Logger ()
        settings = ControllerSettings (idleTimeout=1, timeout=60)
        controller = SinglePageController (url=url, logger=logger,
                service=Process (), behavior=[], settings=settings)
        before = loop.time ()
        await asyncio.wait_for (controller.run (), settings.timeout*2)
        after = loop.time ()
        assert settings.idleTimeout <= after-before <= settings.idleTimeout*2+3
    finally:
        await runner.cleanup ()

def test_set_entry ():
    a = SetEntry (1, a=2, b=3)
    assert a == a
    assert hash (a) == hash (a)

    b = SetEntry (1, a=2, b=4)
    assert a == b
    assert hash (a) == hash (b)

    c = SetEntry (2, a=2, b=3)
    assert a != c
    assert hash (a) != hash (c)

@pytest.mark.asyncio
async def test_idle_state_tracker ():
    # default is idle
    loop = asyncio.get_event_loop ()
    idle = IdleStateTracker (loop)
    assert idle._idle

    # idle change
    await idle.push (PageIdle (False))
    assert not idle._idle

    # nothing happens for other objects
    await idle.push ({})
    assert not idle._idle

    # no state change -> wait does not return
    with pytest.raises (asyncio.TimeoutError):
        await asyncio.wait_for (idle.wait (0.1), timeout=1)

    # wait at least timeout
    delta = 0.2
    timeout = 1
    await idle.push (PageIdle (True))
    assert idle._idle
    start = loop.time ()
    await idle.wait (timeout)
    end = loop.time ()
    assert (timeout-delta) < (end-start) < (timeout+delta)




# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pytest
from .irc import ArgparseBot, RefCountEvent, User, NickMode

def test_mode_parse ():
    assert ArgparseBot.parseMode ('+a') == [('+', 'a')]
    assert ArgparseBot.parseMode ('+ab') == [('+', 'a'), ('+', 'b')]
    assert ArgparseBot.parseMode ('+a+b') == [('+', 'a'), ('+', 'b')]
    assert ArgparseBot.parseMode ('-a') == [('-', 'a')]
    assert ArgparseBot.parseMode ('-ab') == [('-', 'a'), ('-', 'b')]
    assert ArgparseBot.parseMode ('-a-b') == [('-', 'a'), ('-', 'b')]
    assert ArgparseBot.parseMode ('+a-b') == [('+', 'a'), ('-', 'b')]
    assert ArgparseBot.parseMode ('-a+b') == [('-', 'a'), ('+', 'b')]
    assert ArgparseBot.parseMode ('-ab+cd') == [('-', 'a'), ('-', 'b'), ('+', 'c'), ('+', 'd')]

@pytest.fixture
def event ():
    return RefCountEvent ()

def test_refcountevent_arm (event):
    event.arm ()
    assert event.event.is_set ()

def test_refcountevent_ctxmgr (event):
    with event:
        assert event.count == 1
        with event:
            assert event.count == 2

def test_refcountevent_arm_with (event):
    with event:
        event.arm ()
        assert not event.event.is_set ()
    assert event.event.is_set ()

def test_nick_mode ():
    a = User.fromName ('a')
    a2 = User.fromName ('a')
    a3 = User.fromName ('+a')
    b = User.fromName ('+b')
    c = User.fromName ('@c')

    # equality is based on name only, not mode
    assert a == a2
    assert a == a3
    assert a != b

    assert a.hasPriv (None) and not a.hasPriv (NickMode.voice) and not a.hasPriv (NickMode.operator)
    assert b.hasPriv (None) and b.hasPriv (NickMode.voice) and not b.hasPriv (NickMode.operator)
    assert c.hasPriv (None) and c.hasPriv (NickMode.voice) and c.hasPriv (NickMode.operator)




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Communication with Google Chrome through its DevTools protocol.
"""

import json, asyncio, logging, os
from tempfile import mkdtemp
import shutil
import aiohttp, websockets
from yarl import URL

from .util import StrJsonEncoder

logger = logging.getLogger (__name__)

class Browser:
    """
    Communicate with Google Chrome through its DevTools protocol.
    
    Asynchronous context manager that creates a new Tab when entering.
    Destroyed upon exit.
    """

    __slots__ = ('session', 'url', 'tab', 'loop')

    def __init__ (self, url, loop=None):
        self.url = URL (url)
        self.session = None
        self.tab = None
        self.loop = loop

    async def __aiter__ (self):
        """ List all tabs """
        async with aiohttp.ClientSession (loop=self.loop) as session:
            async with session.get (self.url.with_path ('/json/list')) as r:
                resp = await r.json ()
                for tab in resp:
                    if tab['type'] == 'page':
                        yield tab

    async def __aenter__ (self):
        """ Create tab """
        assert self.tab is None
        assert self.session is None
        self.session = aiohttp.ClientSession (loop=self.loop)
        async with self.session.get (self.url.with_path ('/json/new')) as r:
            resp = await r.json ()
            self.tab = await Tab.create (**resp)
            return self.tab

    async def __aexit__ (self, *args):
        assert self.tab is not None
        assert self.session is not None
        await self.tab.close ()
        async with self.session.get (self.url.with_path (f'/json/close/{self.tab.id}')) as r:
            resp = await r.text ()
            assert resp == 'Target is closing'
        self.tab = None
        await self.session.close ()
        self.session = None
        return False

class TabFunction:
    """
    Helper class for infinite-depth tab functions.

    A method usually consists of namespace (Page, Network, …) and function name
    (getFoobar) separated by a dot. This class creates these function names
    while providing an intuitive Python interface (tab.Network.getFoobar).

    This was inspired by pychrome.
    """

    __slots__ = ('name', 'tab')

    def __init__ (self, name, tab):
        self.name = name
        self.tab = tab

    def __eq__ (self, b):
        assert isinstance (b, TabFunction)
        return self.name == b.name

    def __hash__ (self):
        return hash (self.name)

    def __getattr__ (self, k):
        return TabFunction (f'{self.name}.{k}', self.tab)

    async def __call__ (self, **kwargs):
        return await self.tab (self.name, **kwargs)

    def __repr__ (self):
        return f'<TabFunction {self.name}>'

class TabException (Exception):
    pass

class Crashed (TabException):
    pass

class MethodNotFound (TabException):
    pass

class InvalidParameter (TabException):
    pass

# map error codes to native exceptions
errorMap = {-32601: MethodNotFound, -32602: InvalidParameter}

class Tab:
    """
    Communicate with a single Google Chrome browser tab.
    """
    __slots__ = ('id', 'wsUrl', 'ws', 'msgid', 'transactions', 'queue', '_recvHandle', 'crashed')

    def __init__ (self, tabid, ws):
        """ Do not use this method, use Browser context manager. """
        self.id = tabid
        self.ws = ws
        self.msgid = 1
        self.crashed = False
        self.transactions = {}
        self.queue = asyncio.Queue ()

    def __getattr__ (self, k):
        return TabFunction (k, self)

    async def __call__ (self, method, **kwargs):
        """
        Actually call browser method with kwargs
        """

        if self.crashed or self._recvHandle.done ():
            raise Crashed ()

        msgid = self.msgid
        self.msgid += 1
        message = {'method': method, 'params': kwargs, 'id': msgid}
        t = self.transactions[msgid] = {'event': asyncio.Event (), 'result': None}
        logger.debug (f'← {message}')
        await self.ws.send (json.dumps (message, cls=StrJsonEncoder))
        await t['event'].wait ()
        ret = t['result']
        del self.transactions[msgid]
        if isinstance (ret, Exception):
            raise ret
        return ret

    async def _recvProcess (self):
        """
        Receive process that dispatches received websocket frames

        These are either events which will be put into a queue or request
        responses which unblock a __call__.
        """

        async def markCrashed (reason):
            # all pending requests can be considered failed since the
            # browser state is lost
            for v in self.transactions.values ():
                v['result'] = Crashed (reason)
                v['event'].set ()
            # and all future requests will fail as well until reloaded
            self.crashed = True
            await self.queue.put (Crashed (reason))

        while True:
            try:
                msg = await self.ws.recv ()
                msg = json.loads (msg)
            except Exception as e:
                # right now we cannot recover from this
                await markCrashed (e)
                break
            logger.debug (f'→ {msg}')
            if 'id' in msg:
                msgid = msg['id']
                t = self.transactions.get (msgid, None)
                if t is not None:
                    if 'error' in msg:
                        e = msg['error']
                        t['result'] = errorMap.get (e['code'], TabException) (e['code'], e['message'])
                    else:
                        t['result'] = msg['result']
                    t['event'].set ()
                else:
                    # ignore stale result
                    pass # pragma: no cover
            elif 'method' in msg:
                # special treatment
                if msg['method'] == 'Inspector.targetCrashed':
                    await markCrashed ('target')
                else:
                    await self.queue.put (msg)
            else:
                assert False # pragma: no cover

    async def run (self):
        self._recvHandle = asyncio.ensure_future (self._recvProcess ())

    async def close (self):
        self._recvHandle.cancel ()
        await self.ws.close ()
        # no join, throw away the queue. There will be nobody listening on the
        # other end.
        #await self.queue.join ()

    @property
    def pending (self):
        return self.queue.qsize ()

    async def get (self):
        def getattrRecursive (obj, name):
            if '.' in name:
                n, ext = name.split ('.', 1)
                return getattrRecursive (getattr (obj, n), ext)
            return getattr (obj, name)

        if self.crashed:
            raise Crashed ()

        ret = await self.queue.get ()
        if isinstance (ret, Exception):
            raise ret
        return getattrRecursive (self, ret['method']), ret['params']

    @classmethod
    async def create (cls, **kwargs):
        """ Async init """
        # increase size limit of a single frame to something ridiciously high,
        # so we can safely grab screenshots
        maxSize = 100*1024*1024 # 100 MB
        # chrome does not like pings and kills the connection, disable them
        ws = await websockets.connect(kwargs['webSocketDebuggerUrl'],
                max_size=maxSize, ping_interval=None)
        ret = cls (kwargs['id'], ws)
        await ret.run ()
        return ret

class Process:
    """ Start Google Chrome listening on a random port """

    __slots__ = ('binary', 'windowSize', 'p', 'userDataDir')

    def __init__ (self, binary='google-chrome-stable', windowSize=(1920, 1080)):
        self.binary = binary
        self.windowSize = windowSize
        self.p = None

    async def __aenter__ (self):
        assert self.p is None
        self.userDataDir = mkdtemp (prefix=__package__ + '-chrome-userdata-')
        # see https://github.com/GoogleChrome/chrome-launcher/blob/master/docs/chrome-flags-for-tools.md
        args = [self.binary,
                '--window-size={},{}'.format (*self.windowSize),
                f'--user-data-dir={self.userDataDir}', # use temporory user dir
                '--no-default-browser-check',
                '--no-first-run', # don’t show first run screen
                '--disable-breakpad', # no error reports
                '--disable-extensions',
                '--disable-infobars',
                '--disable-notifications', # no libnotify
                '--disable-background-networking', # disable background services (updating, safe browsing, …)
                '--safebrowsing-disable-auto-update',
                '--disable-sync', # no google account syncing
                '--metrics-recording-only', # do not submit metrics
                '--disable-default-apps',
                '--disable-background-timer-throttling',
                '--disable-client-side-phishing-detection',
                '--disable-popup-blocking',
                '--disable-prompt-on-repost',
                '--enable-automation', # enable various automation-related things
                '--password-store=basic',
                '--headless',
                '--disable-gpu',
                '--hide-scrollbars', # hide scrollbars on screenshots
                '--mute-audio', # don’t play any audio
                '--remote-debugging-port=0', # pick a port. XXX: we may want to use --remote-debugging-pipe instead
                '--homepage=about:blank',
                'about:blank']
        # start new session, so ^C does not affect subprocess
        self.p = await asyncio.create_subprocess_exec (*args,
                stdout=asyncio.subprocess.DEVNULL,
                stderr=asyncio.subprocess.DEVNULL,
                stdin=asyncio.subprocess.DEVNULL,
                start_new_session=True)
        port = None
        # chrome writes its current active devtools port to a file. due to the
        # sleep() this is rather ugly, but should work with all versions of the
        # browser.
        for i in range (100):
            try:
                with open (os.path.join (self.userDataDir, 'DevToolsActivePort'), 'r') as fd:
                    port = int (fd.readline ().strip ())
                    break
            except FileNotFoundError:
                await asyncio.sleep (0.2)
        if port is None:
            raise Exception ('Chrome died on us.')

        return URL.build(scheme='http', host='localhost', port=port)

    async def __aexit__ (self, *exc):
        self.p.terminate ()
        await self.p.wait ()

        # Try to delete the temporary directory multiple times. It looks like
        # Chrome will change files in there even after it exited (i.e. .wait()
        # returned). Very strange.
        for i in range (5):
            try:
                shutil.rmtree (self.userDataDir)
                break
            except:
                await asyncio.sleep (0.2)

        self.p = None
        return False

class Passthrough:
    __slots__ = ('url', )

    def __init__ (self, url):
        self.url = URL (url)

    async def __aenter__ (self):
        return self.url

    async def __aexit__ (self, *exc):
        return False




# Copyright (c) 2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Misc tools
"""

import shutil, sys, os, logging, argparse, json
from io import BytesIO

from warcio.archiveiterator import ArchiveIterator
from warcio.warcwriter import WARCWriter
from yarl import URL

from pkg_resources import parse_version, parse_requirements

from .util import getSoftwareInfo, StrJsonEncoder

def mergeWarc (files, output):
    # stats
    unique = 0
    revisit = 0
    uniqueLength = 0
    revisitLength = 0

    payloadMap = {}
    writer = WARCWriter (output, gzip=True)

    # Add an additional warcinfo record, describing the transformations. This
    # is not ideal, since
    #   “A ‘warcinfo’ record describes the records that
    #   follow it […] until next ‘warcinfo’”
    #   -- https://iipc.github.io/warc-specifications/specifications/warc-format/warc-1.1/#warcinfo
    # A warcinfo record is expected at the beginning of every file. But it
    # might have written by a different software, so we don’t want to
    # strip/replace that information, but supplement it.
    warcinfo = {
            'software': getSoftwareInfo (),
            'tool': 'crocoite-merge', # not the name of the cli tool
            'parameters': {'inputs': files},
            }
    payload = BytesIO (json.dumps (warcinfo, indent=2).encode ('utf-8'))
    record = writer.create_warc_record ('', 'warcinfo',
            payload=payload,
            warc_headers_dict={'Content-Type': 'text/plain; encoding=utf-8'})
    writer.write_record (record)

    for l in files:
        with open (l, 'rb') as fd:
            for record in ArchiveIterator (fd):
                if record.rec_type in {'resource', 'response'}:
                    headers = record.rec_headers
                    rid = headers.get_header('WARC-Record-ID')
                    csum = headers.get_header('WARC-Payload-Digest')
                    length = int (headers.get_header ('Content-Length'))
                    dup = payloadMap.get (csum, None)
                    if dup is None:
                        payloadMap[csum] = {'uri': headers.get_header('WARC-Target-URI'),
                                'id': rid, 'date': headers.get_header('WARC-Date')}
                        unique += 1
                        uniqueLength += length
                    else:
                        logging.debug (f'Record {rid} is duplicate of {dup["id"]}')
                        # Payload may be identical, but HTTP headers are
                        # (probably) not. Include them.
                        record = writer.create_revisit_record (
                                headers.get_header('WARC-Target-URI'), digest=csum,
                                refers_to_uri=dup['uri'], refers_to_date=dup['date'],
                                http_headers=record.http_headers)
                        record.rec_headers.add_header ('WARC-Truncated', 'length')
                        record.rec_headers.add_header ('WARC-Refers-To', dup['id'])
                        revisit += 1
                        revisitLength += length
                else:
                    unique += 1
                writer.write_record (record)
    json.dump (dict (
            unique=dict (records=unique, bytes=uniqueLength),
            revisit=dict (records=revisit, bytes=revisitLength),
            ratio=dict (
                    records=unique/(unique+revisit),
                    bytes=uniqueLength/(uniqueLength+revisitLength)
                    ),
            ),
            sys.stdout,
            cls=StrJsonEncoder)
    sys.stdout.write ('\n')

def mergeWarcCli():
    parser = argparse.ArgumentParser(description='Merge WARCs, reads filenames from stdin.')
    parser.add_argument('--verbose', '-v', action='store_true')
    parser.add_argument('output', type=argparse.FileType ('wb'), help='Output WARC')

    args = parser.parse_args()
    loglevel = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig (level=loglevel)

    mergeWarc([l.strip() for l in sys.stdin], args.output)

def extractScreenshot ():
    """
    Extract page screenshots from a WARC generated by crocoite into files
    """

    parser = argparse.ArgumentParser(description='Extract screenshots from '
            'WARC, write JSON info to stdout.')
    parser.add_argument('-f', '--force', action='store_true',
            help='Overwrite existing files')
    parser.add_argument('-1', '--one', action='store_true',
            help='Only extract the first screenshot into a file named prefix')
    parser.add_argument('input', type=argparse.FileType ('rb'),
            help='Input WARC')
    parser.add_argument('prefix', help='Output file prefix')

    args = parser.parse_args()

    i = 0
    with args.input:
        for record in ArchiveIterator (args.input):
            headers = record.rec_headers
            if record.rec_type != 'conversion' or \
                    headers['Content-Type'] != 'image/png' or \
                    'X-Crocoite-Screenshot-Y-Offset' not in headers:
                continue

            url = URL (headers.get_header ('WARC-Target-URI'))
            yoff = int (headers.get_header ('X-Crocoite-Screenshot-Y-Offset'))
            outpath = f'{args.prefix}{i:05d}.png' if not args.one else args.prefix
            if args.force or not os.path.exists (outpath):
                json.dump ({'file': outpath, 'url': url, 'yoff': yoff},
                        sys.stdout, cls=StrJsonEncoder)
                sys.stdout.write ('\n')
                with open (outpath, 'wb') as out:
                    shutil.copyfileobj (record.raw_stream, out)
                i += 1
            else:
                print (f'not overwriting {outpath}', file=sys.stderr)

            if args.one:
                break

class Errata:
    __slots__ = ('uuid', 'description', 'affects')

    def __init__ (self, uuid, description, affects):
        self.uuid = uuid
        self.description = description
        # slightly abusing setuptool’s version parsing/matching here
        self.affects = list (parse_requirements(affects))

    def __contains__ (self, pkg):
        """
        Return True if the versions in pkg are affected by this errata

        pkg must be a mapping from project_name to version
        """
        matchedAll = []
        for a in self.affects:
            haveVersion = pkg.get (a.project_name, None)
            matchedAll.append (haveVersion is not None and haveVersion in a)
        return all (matchedAll)

    def __repr__ (self):
        return f'{self.__class__.__name__}({self.uuid!r}, {self.description!r}, {self.affects!r})'

    @property
    def fixable (self):
        return getattr (self, 'applyFix', None) is not None

    def toDict (self):
        return {'uuid': self.uuid,
                'description': self.description,
                'affects': list (map (str, self.affects)),
                'fixable': self.fixable}

class FixableErrata(Errata):
    def applyFix (self, records):
        raise NotImplementedError () # pragma: no cover

bugs = [
    Errata (uuid='34a176b3-ad3d-430f-a082-68087f304572',
            description='Generated by version < 1.0. No erratas are supported for this version.',
            affects=['crocoite<1.0'],
            ),
    ]

def makeReport (fd):
    for record in ArchiveIterator (fd):
        if record.rec_type == 'warcinfo':
            try:
                data = json.load (record.raw_stream)
                haveVersions = dict ([(pkg['projectName'], parse_version(pkg['version'])) for pkg in data['software']['self']])
                yield from filter (lambda b: haveVersions in b, bugs)
            except json.decoder.JSONDecodeError:
                pass

def errata ():
    parser = argparse.ArgumentParser(description=f'Show/fix erratas for WARCs generated by {__package__}.')
    parser.add_argument('input', type=argparse.FileType ('rb'), help='Input WARC')

    args = parser.parse_args()

    hasErrata = False
    for item in makeReport (args.input):
        json.dump (item.toDict (), sys.stdout)
        sys.stdout.write ('\n')
        sys.stdout.flush ()
        hasErrata = True
    return int (hasErrata)




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Chrome browser interactions.
"""

import asyncio
from base64 import b64decode, b64encode
from datetime import datetime, timedelta
from http.server import BaseHTTPRequestHandler

from yarl import URL
from multidict import CIMultiDict

from .logger import Level
from .devtools import Browser, TabException

# These two classes’ only purpose is so we can later tell whether a body was
# base64-encoded or a unicode string
class Base64Body (bytes):
    def __new__ (cls, value):
        return bytes.__new__ (cls, b64decode (value))

    @classmethod
    def fromBytes (cls, b):
        """ For testing """
        return cls (b64encode (b))

class UnicodeBody (bytes):
    def __new__ (cls, value):
        if type (value) is not str:
            raise TypeError ('expecting unicode string')

        return bytes.__new__ (cls, value.encode ('utf-8'))

class Request:
    __slots__ = ('headers', 'body', 'initiator', 'hasPostData', 'method', 'timestamp')

    def __init__ (self, method=None, headers=None, body=None):
        self.headers = headers
        self.body = body
        self.hasPostData = False
        self.initiator = None
        # HTTP method
        self.method = method
        self.timestamp = None

    def __repr__ (self):
        return f'Request({self.method!r}, {self.headers!r}, {self.body!r})'

    def __eq__ (self, b):
        if b is None:
            return False

        if not isinstance (b, Request):
            raise TypeError ('Can only compare equality with Request.')

        # do not compare hasPostData (only required to fetch body) and
        # timestamp (depends on time)
        return self.headers == b.headers and \
                self.body == b.body and \
                self.initiator == b.initiator and \
                self.method == b.method

class Response:
    __slots__ = ('status', 'statusText', 'headers', 'body', 'bytesReceived',
            'timestamp', 'mimeType')

    def __init__ (self, status=None, statusText=None, headers=None, body=None, mimeType=None):
        self.status = status
        self.statusText = statusText
        self.headers = headers
        self.body = body
        # bytes received over the network (not body size!)
        self.bytesReceived = 0
        self.timestamp = None
        self.mimeType = mimeType

    def __repr__ (self):
        return f'Response({self.status!r}, {self.statusText!r}, {self.headers!r}, {self.body!r}, {self.mimeType!r})'

    def __eq__ (self, b):
        if b is None:
            return False

        if not isinstance (b, Response):
            raise TypeError ('Can only compare equality with Response.')

        # do not compare bytesReceived (depends on network), timestamp
        # (depends on time) and statusText (does not matter)
        return self.status == b.status and \
                self.statusText == b.statusText and \
                self.headers == b.headers and \
                self.body == b.body and \
                self.mimeType == b.mimeType

class ReferenceTimestamp:
    """ Map relative timestamp to absolute timestamp """

    def __init__ (self, relative, absolute):
        self.relative = timedelta (seconds=relative)
        self.absolute = datetime.utcfromtimestamp (absolute)

    def __call__ (self, relative):
        if not isinstance (relative, timedelta):
            relative = timedelta (seconds=relative)
        return self.absolute + (relative-self.relative)

class RequestResponsePair:
    __slots__ = ('request', 'response', 'id', 'url', 'remoteIpAddress',
            'protocol', 'resourceType', '_time')

    def __init__ (self, id=None, url=None, request=None, response=None):
        self.request = request
        self.response = response
        self.id = id
        self.url = url
        self.remoteIpAddress = None
        self.protocol = None
        self.resourceType = None
        self._time = None

    def __repr__ (self):
        return f'RequestResponsePair({self.id!r}, {self.url!r}, {self.request!r}, {self.response!r})'

    def __eq__ (self, b):
        if not isinstance (b, RequestResponsePair):
            raise TypeError (f'Can only compare with {self.__class__.__name__}')

        # do not compare id and _time. These depend on external factors and do
        # not influence the request/response *content*
        return self.request == b.request and \
                self.response == b.response and \
                self.url == b.url and \
                self.remoteIpAddress == b.remoteIpAddress and \
                self.protocol == b.protocol and \
                self.resourceType == b.resourceType

    def fromRequestWillBeSent (self, req):
        """ Set request data from Chrome Network.requestWillBeSent event """
        r = req['request']

        self.id = req['requestId']
        self.url = URL (r['url'])
        self.resourceType = req.get ('type')
        self._time = ReferenceTimestamp (req['timestamp'], req['wallTime'])

        assert self.request is None, req
        self.request = Request ()
        self.request.initiator = req['initiator']
        self.request.headers = CIMultiDict (self._unfoldHeaders (r['headers']))
        self.request.hasPostData = r.get ('hasPostData', False)
        self.request.method = r['method']
        self.request.timestamp = self._time (req['timestamp'])
        if self.request.hasPostData:
            postData = r.get ('postData')
            if postData is not None:
                self.request.body = UnicodeBody (postData)

    def fromResponse (self, r, timestamp=None, resourceType=None):
        """
        Set response data from Chrome’s Response object.
        
        Request must exist. Updates if response was set before. Sometimes
        fromResponseReceived is triggered twice by Chrome. No idea why.
        """
        assert self.request is not None, (self.request, r)

        if not timestamp:
            timestamp = self.request.timestamp

        self.remoteIpAddress = r.get ('remoteIPAddress')
        self.protocol = r.get ('protocol')
        if resourceType:
            self.resourceType = resourceType

        # a response may contain updated request headers (i.e. those actually
        # sent over the wire)
        if 'requestHeaders' in r:
            self.request.headers = CIMultiDict (self._unfoldHeaders (r['requestHeaders']))

        self.response = Response ()
        self.response.headers = CIMultiDict (self._unfoldHeaders (r['headers']))
        self.response.status = r['status']
        self.response.statusText = r['statusText']
        self.response.timestamp = timestamp
        self.response.mimeType = r['mimeType']

    def fromResponseReceived (self, resp):
        """ Set response data from Chrome Network.responseReceived """
        return self.fromResponse (resp['response'],
                self._time (resp['timestamp']), resp['type'])

    def fromLoadingFinished (self, data):
        self.response.bytesReceived = data['encodedDataLength']

    def fromLoadingFailed (self, data):
        self.response = None

    @staticmethod
    def _unfoldHeaders (headers):
        """
        A host may send multiple headers using the same key, which Chrome folds
        into the same item. Separate those.
        """
        items = []
        for k in headers.keys ():
            for v in headers[k].split ('\n'):
                items.append ((k, v))
        return items

    async def prefetchRequestBody (self, tab):
        if self.request.hasPostData and self.request.body is None:
            try:
                postData = await tab.Network.getRequestPostData (requestId=self.id)
                self.request.body = UnicodeBody (postData['postData'])
            except TabException:
                self.request.body = None

    async def prefetchResponseBody (self, tab):
        """ Fetch response body """
        try:
            body = await tab.Network.getResponseBody (requestId=self.id)
            if body['base64Encoded']:
                self.response.body = Base64Body (body['body'])
            else:
                self.response.body = UnicodeBody (body['body'])
        except TabException:
            self.response.body = None

class NavigateError (IOError):
    pass

class PageIdle:
    """ Page idle event """

    __slots__ = ('idle', )

    def __init__ (self, idle):
        self.idle = idle

    def __bool__ (self):
        return self.idle

class FrameNavigated:
    __slots__ = ('id', 'url', 'mimeType')

    def __init__ (self, id, url, mimeType):
        self.id = id
        self.url = URL (url)
        self.mimeType = mimeType

class SiteLoader:
    """
    Load site in Chrome and monitor network requests

    XXX: track popup windows/new tabs and close them
    """

    __slots__ = ('requests', 'browser', 'logger', 'tab', '_iterRunning',
            '_framesLoading', '_rootFrame')
    allowedSchemes = {'http', 'https'}

    def __init__ (self, browser, logger):
        self.requests = {}
        self.browser = Browser (url=browser)
        self.logger = logger.bind (context=type (self).__name__)
        self._iterRunning = []

        self._framesLoading = set ()
        self._rootFrame = None

    async def __aenter__ (self):
        tab = self.tab = await self.browser.__aenter__ ()

        # enable events
        await asyncio.gather (*[
                tab.Log.enable (),
                tab.Network.enable(),
                tab.Page.enable (),
                tab.Inspector.enable (),
                tab.Network.clearBrowserCache (),
                tab.Network.clearBrowserCookies (),
                ])
        return self

    async def __aexit__ (self, exc_type, exc_value, traceback):
        for task in self._iterRunning:
            # ignore any results from stuff we did not end up using anyway
            if not task.done ():
                task.cancel ()
        self._iterRunning = []
        await self.browser.__aexit__ (exc_type, exc_value, traceback)
        self.tab = None
        return False

    def __len__ (self):
        return len (self.requests)

    async def __aiter__ (self):
        """ Retrieve network items """
        tab = self.tab
        assert tab is not None
        handler = {
                tab.Network.requestWillBeSent: self._requestWillBeSent,
                tab.Network.responseReceived: self._responseReceived,
                tab.Network.loadingFinished: self._loadingFinished,
                tab.Network.loadingFailed: self._loadingFailed,
                tab.Log.entryAdded: self._entryAdded,
                tab.Page.javascriptDialogOpening: self._javascriptDialogOpening,
                tab.Page.frameStartedLoading: self._frameStartedLoading,
                tab.Page.frameStoppedLoading: self._frameStoppedLoading,
                tab.Page.frameNavigated: self._frameNavigated,
                }

        # The implementation is a little advanced. Why? The goal here is to
        # process events from the tab as quickly as possible (i.e.
        # asynchronously). We need to make sure that JavaScript dialogs are
        # handled immediately for instance. Otherwise they stall every
        # other request. Also, we don’t want to use an unbounded queue,
        # since the items yielded can get quite big (response body). Thus
        # we need to block (yield) for every item completed, but not
        # handled by the consumer (caller).
        running = self._iterRunning
        tabGetTask = asyncio.ensure_future (self.tab.get ())
        running.append (tabGetTask)
        while True:
            done, pending = await asyncio.wait (running, return_when=asyncio.FIRST_COMPLETED)
            for t in done:
                result = t.result ()
                if result is None:
                    pass
                elif t == tabGetTask:
                    method, data = result
                    f = handler.get (method, None)
                    if f is not None:
                        task = asyncio.ensure_future (f (**data))
                        pending.add (task)
                    tabGetTask = asyncio.ensure_future (self.tab.get ())
                    pending.add (tabGetTask)
                else:
                    yield result

            running = pending
            self._iterRunning = running

    async def navigate (self, url):
        ret = await self.tab.Page.navigate(url=url)
        self.logger.debug ('navigate',
                uuid='9d47ded2-951f-4e09-86ee-fd4151e20666', result=ret)
        if 'errorText' in ret:
            raise NavigateError (ret['errorText'])
        self._rootFrame = ret['frameId']

    # internal chrome callbacks
    async def _requestWillBeSent (self, **kwargs):
        self.logger.debug ('requestWillBeSent',
                uuid='b828d75a-650d-42d2-8c66-14f4547512da', args=kwargs)

        reqId = kwargs['requestId']
        req = kwargs['request']
        url = URL (req['url'])
        logger = self.logger.bind (reqId=reqId, reqUrl=url)

        if url.scheme not in self.allowedSchemes:
            return

        ret = None
        item = self.requests.get (reqId)
        if item:
            # redirects never “finish” loading, but yield another requestWillBeSent with this key set
            redirectResp = kwargs.get ('redirectResponse')
            if redirectResp:
                if item.url != url:
                    # this happens for unknown reasons. the docs simply state
                    # it can differ in case of a redirect. Fix it and move on.
                    logger.warning ('redirect url differs',
                            uuid='558a7df7-2258-4fe4-b16d-22b6019cc163',
                            expected=item.url)
                    redirectResp['url'] = str (item.url)
                item.fromResponse (redirectResp)
                logger.info ('redirect', uuid='85eaec41-e2a9-49c2-9445-6f19690278b8', target=url)
                # XXX: queue this? no need to wait for it
                await item.prefetchRequestBody (self.tab)
                # cannot fetch response body due to race condition (item id reused)
                ret = item
            else:
                logger.warning ('request exists', uuid='2c989142-ba00-4791-bb03-c2a14e91a56b')

        item = RequestResponsePair ()
        item.fromRequestWillBeSent (kwargs)
        self.requests[reqId] = item

        return ret

    async def _responseReceived (self, **kwargs):
        self.logger.debug ('responseReceived',
                uuid='ecd67e69-401a-41cb-b4ec-eeb1f1ec6abb', args=kwargs)

        reqId = kwargs['requestId']
        item = self.requests.get (reqId)
        if item is None:
            return

        resp = kwargs['response']
        url = URL (resp['url'])
        logger = self.logger.bind (reqId=reqId, respUrl=url)
        if item.url != url:
            logger.error ('url mismatch', uuid='7385f45f-0b06-4cbc-81f9-67bcd72ee7d0', respUrl=url)
        if url.scheme in self.allowedSchemes:
            item.fromResponseReceived (kwargs)
        else:
            logger.warning ('scheme forbidden', uuid='2ea6e5d7-dd3b-4881-b9de-156c1751c666')

    async def _loadingFinished (self, **kwargs):
        """
        Item was fully loaded. For some items the request body is not available
        when responseReceived is fired, thus move everything here.
        """
        self.logger.debug ('loadingFinished',
                uuid='35479405-a5b5-4395-8c33-d3601d1796b9', args=kwargs)

        reqId = kwargs['requestId']
        item = self.requests.pop (reqId, None)
        if item is None:
            # we never recorded this request (blacklisted scheme, for example)
            return
        req = item.request
        if item.url.scheme in self.allowedSchemes:
            item.fromLoadingFinished (kwargs)
            # XXX queue both
            await asyncio.gather (item.prefetchRequestBody (self.tab), item.prefetchResponseBody (self.tab))
            return item

    async def _loadingFailed (self, **kwargs):
        self.logger.info ('loadingFailed',
                uuid='4a944e85-5fae-4aa6-9e7c-e578b29392e4', args=kwargs)

        reqId = kwargs['requestId']
        logger = self.logger.bind (reqId=reqId)
        item = self.requests.pop (reqId, None)
        if item is not None:
            item.fromLoadingFailed (kwargs)
            return item

    async def _entryAdded (self, **kwargs):
        """ Log entry added """
        entry = kwargs['entry']
        level = {'verbose': Level.DEBUG, 'info': Level.INFO,
                'warning': Level.WARNING,
                'error': Level.ERROR}.get (entry.pop ('level'), Level.INFO)
        entry['uuid'] = 'e62ffb5a-0521-459c-a3d9-1124551934d2'
        self.logger (level, 'console', **entry)

    async def _javascriptDialogOpening (self, **kwargs):
        t = kwargs.get ('type')
        if t in {'alert', 'confirm', 'prompt'}:
            self.logger.info ('js dialog',
                    uuid='d6f07ce2-648e-493b-a1df-f353bed27c84',
                    action='cancel', type=t, message=kwargs.get ('message'))
            await self.tab.Page.handleJavaScriptDialog (accept=False)
        elif t == 'beforeunload':
            # we must accept this one, otherwise the page will not unload/close
            self.logger.info ('js dialog',
                    uuid='96399b99-9834-4c8f-bd93-cb9fa2225abd',
                    action='proceed', type=t, message=kwargs.get ('message'))
            await self.tab.Page.handleJavaScriptDialog (accept=True)
        else: # pragma: no cover
            self.logger.warning ('js dialog unknown',
                    uuid='3ef7292e-8595-4e89-b834-0cc6bc40ee38', **kwargs)

    async def _frameStartedLoading (self, **kwargs):
        self.logger.debug ('frameStartedLoading',
                uuid='bbeb39c0-3304-4221-918e-f26bd443c566', args=kwargs)

        self._framesLoading.add (kwargs['frameId'])
        return PageIdle (False)

    async def _frameStoppedLoading (self, **kwargs):
        self.logger.debug ('frameStoppedLoading',
                uuid='fcbe8110-511c-4cbb-ac2b-f61a5782c5a0', args=kwargs)

        self._framesLoading.remove (kwargs['frameId'])
        if not self._framesLoading:
            return PageIdle (True)

    async def _frameNavigated (self, **kwargs):
        self.logger.debug ('frameNavigated',
                uuid='0e876f7d-7129-4612-8632-686f42ac6e1f', args=kwargs)
        frame = kwargs['frame']
        if self._rootFrame == frame['id']:
            assert frame.get ('parentId', None) is None, "root frame must not have a parent"
            return FrameNavigated (frame['id'], frame['url'], frame['mimeType'])




# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Behavior scripts (i.e. subclasses of Behavior) are a powerful method to
manipulate websites loaded into Chrome. They are executed by the controller
after the page started loading (onload), after it has been idle for a while
(onstop) and after loading was stopped (onfinish).

The script’s excercise their power either through DevTools API calls or by
injecting JavaScript into the page context. Thus they can manipulate both, the
browser itself (DevTools; modify resolution, get DOM snapshot) as well as the
page (JavaScript; trigger JavaScript events, call web API’s).

They also emit (yield) data processable by any consumer registered to the
controller. This allows storing captured screenshots inside WARC files, for
instance.
"""

import asyncio, json, os.path
from base64 import b64decode
from collections import OrderedDict
import pkg_resources

from html5lib.serializer import HTMLSerializer
from yarl import URL
import yaml

from .util import getFormattedViewportMetrics
from . import html
from .html import StripAttributeFilter, StripTagFilter, ChromeTreeWalker
from .devtools import Crashed, TabException

class Script:
    """ A JavaScript resource """

    __slots__ = ('path', 'data')
    datadir = 'data'

    def __init__ (self, path=None, encoding='utf-8'):
        self.path = path
        if path:
            self.data = pkg_resources.resource_string (__name__, os.path.join (self.datadir, path)).decode (encoding)

    def __repr__ (self):
        return f'<Script {self.path}>'

    def __str__ (self):
        return self.data

    @property
    def abspath (self):
        return pkg_resources.resource_filename (__name__,
                os.path.join (self.datadir, self.path))

    @classmethod
    def fromStr (cls, data, path=None):
        s = Script ()
        s.data = data
        s.path = path
        return s

class Behavior:
    __slots__ = ('loader', 'logger')

    # unique behavior name
    name = None

    def __init__ (self, loader, logger):
        assert self.name is not None
        self.loader = loader
        self.logger = logger.bind (context=type (self).__name__)

    def __contains__ (self, url):
        """
        Accept every URL by default
        """
        return True

    def __repr__ (self):
        return f'<Behavior {self.name}>'

    async def onload (self):
        """ After loading the page started """
        # this is a dirty hack to make this function an async generator
        return
        yield # pragma: no cover

    async def onstop (self):
        """ Before page loading is stopped """
        return
        yield # pragma: no cover

    async def onfinish (self):
        """ After the site has stopped loading """
        return
        yield # pragma: no cover

class JsOnload (Behavior):
    """ Execute JavaScript on page load """

    __slots__ = ('script', 'context', 'options')

    scriptPath = None

    def __init__ (self, loader, logger):
        super ().__init__ (loader, logger)
        self.script = Script (self.scriptPath)
        self.context = None
        # options passed to constructor
        self.options = {}

    async def onload (self):
        tab = self.loader.tab
        yield self.script

        # This is slightly awkward, since we cannot compile the class into an
        # objectId and then reference it. Therefore the script must return a
        # class constructor, which is then called with a generic options
        # parameter.
        # XXX: is there a better way to do this?
        result = await tab.Runtime.evaluate (expression=str (self.script))
        self.logger.debug ('behavior onload inject',
                uuid='a2da9b78-5648-44c5-bfa8-5c7573e13ad3', result=result)
        exception = result.get ('exceptionDetails', None)
        result = result['result']
        assert result['type'] == 'function', result
        assert result.get ('subtype') != 'error', exception
        constructor = result['objectId']

        if self.options:
            yield Script.fromStr (json.dumps (self.options, indent=2), f'{self.script.path}#options')
        result = await tab.Runtime.callFunctionOn (
                functionDeclaration='function(options){return new this(options);}',
                objectId=constructor,
                arguments=[{'value': self.options}])
        self.logger.debug ('behavior onload start',
                uuid='6c0605ae-93b3-46b3-b575-ba45790909a7', result=result)
        result = result['result']
        assert result['type'] == 'object', result
        assert result.get ('subtype') != 'error', result
        self.context = result['objectId']

    async def onstop (self):
        tab = self.loader.tab
        try:
            assert self.context is not None
            await tab.Runtime.callFunctionOn (functionDeclaration='function(){return this.stop();}',
                    objectId=self.context)
            await tab.Runtime.releaseObject (objectId=self.context)
        except TabException as e:
            # cannot do anything about that. Ignoring should be fine.
            self.logger.error ('jsonload onstop failed',
                    uuid='1786726f-c8ec-4f79-8769-30954d4e32f5',
                    exception=e.args,
                    objectId=self.context)

        return
        yield # pragma: no cover

### Generic scripts ###

class Scroll (JsOnload):
    name = 'scroll'
    scriptPath = 'scroll.js'

class EmulateScreenMetrics (Behavior):
    name = 'emulateScreenMetrics'

    async def onstop (self):
        """
        Emulate different screen sizes, causing the site to fetch assets (img
        srcset and css, for example) for different screen resolutions.
        """
        cssPpi = 96
        sizes = [
                {'width': 1920, 'height': 1080, 'deviceScaleFactor': 1.5, 'mobile': False},
                {'width': 1920, 'height': 1080, 'deviceScaleFactor': 2, 'mobile': False},
                # very dense display
                {'width': 1920, 'height': 1080, 'deviceScaleFactor': 4, 'mobile': False},
                # just a few samples:
                # 1st gen iPhone (portrait mode)
                {'width': 320, 'height': 480, 'deviceScaleFactor': 163/cssPpi, 'mobile': True},
                # 6th gen iPhone (portrait mode)
                {'width': 750, 'height': 1334, 'deviceScaleFactor': 326/cssPpi, 'mobile': True},
                ]
        l = self.loader
        tab = l.tab
        for s in sizes:
            self.logger.debug ('device override',
                    uuid='3d2d8096-1a75-4830-ad79-ae5f6f97071d', **s)
            await tab.Emulation.setDeviceMetricsOverride (**s)
            # give the browser time to re-eval page and start requests
            # XXX: should wait until loader is not busy any more
            await asyncio.sleep (1)
        self.logger.debug ('clear override',
                uuid='f9401683-eb3a-4b86-9bb2-c8c5d876fc8d')
        await tab.Emulation.clearDeviceMetricsOverride ()
        return
        yield # pragma: no cover

class DomSnapshotEvent:
    __slots__ = ('url', 'document', 'viewport')

    def __init__ (self, url, document, viewport):
        # XXX: document encoding?
        assert isinstance (document, bytes)

        self.url = url
        self.document = document
        self.viewport = viewport

class DomSnapshot (Behavior):
    """
    Get a DOM snapshot of tab and write it to WARC.

    We could use DOMSnapshot.getSnapshot here, but the API is not stable
    yet. Also computed styles are not really necessary here.
    """

    __slots__ = ('script', )

    name = 'domSnapshot'

    def __init__ (self, loader, logger):
        super ().__init__ (loader, logger)
        self.script = Script ('canvas-snapshot.js')

    async def onfinish (self):
        tab = self.loader.tab

        yield self.script
        await tab.Runtime.evaluate (expression=str (self.script), returnByValue=True)

        viewport = await getFormattedViewportMetrics (tab)
        dom = await tab.DOM.getDocument (depth=-1, pierce=True)
        self.logger.debug ('dom snapshot document',
                uuid='0c720784-8bd1-4fdc-a811-84394d753539', dom=dom)
        haveUrls = set ()
        for doc in ChromeTreeWalker (dom['root']).split ():
            url = URL (doc['documentURL'])
            if url in haveUrls:
                # ignore duplicate URLs. they are usually caused by
                # javascript-injected iframes (advertising) with no(?) src
                self.logger.warning ('dom snapshot duplicate',
                        uuid='d44de989-98d4-456e-82e7-9d4c49acab5e')
            elif url.scheme in ('http', 'https'):
                self.logger.debug ('dom snapshot',
                        uuid='ece7ff05-ccd9-44b5-b6a8-be25a24b96f4',
                        base=doc["baseURL"])
                haveUrls.add (url)
                walker = ChromeTreeWalker (doc)
                # remove script, to make the page static and noscript, because at the
                # time we took the snapshot scripts were enabled
                disallowedTags = ['script', 'noscript']
                disallowedAttributes = html.eventAttributes
                stream = StripAttributeFilter (StripTagFilter (walker, disallowedTags), disallowedAttributes)
                serializer = HTMLSerializer ()
                yield DomSnapshotEvent (url.with_fragment(None), serializer.render (stream, 'utf-8'), viewport)

class ScreenshotEvent:
    __slots__ = ('yoff', 'data', 'url')

    def __init__ (self, url, yoff, data):
        self.url = url
        self.yoff = yoff
        self.data = data

class Screenshot (Behavior):
    """
    Create screenshot from tab and write it to WARC
    """

    __slots__ = ('script')

    name = 'screenshot'

    # Hardcoded max texture size of 16,384 (crbug.com/770769)
    maxDim = 16*1024

    def __init__ (self, loader, logger):
        super ().__init__ (loader, logger)
        self.script = Script ('screenshot.js')

    async def onfinish (self):
        tab = self.loader.tab

        # for top-level/full-screen elements with position: fixed we need to
        # figure out their actual size (i.e. scrollHeight) and use that when
        # overriding the viewport size.
        # we could do this without javascript, but that would require several
        # round-trips to Chrome or pulling down the entire DOM+computed styles
        tab = self.loader.tab
        yield self.script
        result = await tab.Runtime.evaluate (expression=str (self.script), returnByValue=True)
        assert result['result']['type'] == 'object', result
        result = result['result']['value']

        # this is required to make the browser render more than just the small
        # actual viewport (i.e. entire page).  see
        # https://github.com/GoogleChrome/puppeteer/blob/45873ea737b4ebe4fa7d6f46256b2ea19ce18aa7/lib/Page.js#L805
        metrics = await tab.Page.getLayoutMetrics ()
        contentSize = metrics['contentSize']
        contentHeight = max (result + [contentSize['height']])

        override = {
                'width': 0,
                'height': 0,
                'deviceScaleFactor': 0,
                'mobile': False,
                'viewport': {'x': 0,
                    'y': 0,
                    'width': contentSize['width'],
                    'height': contentHeight,
                    'scale': 1}
                }
        self.logger.debug ('screenshot override',
                uuid='e0affa18-cbb1-4d97-9d13-9a88f704b1b2', override=override)
        await tab.Emulation.setDeviceMetricsOverride (**override)

        tree = await tab.Page.getFrameTree ()
        try:
            url = URL (tree['frameTree']['frame']['url']).with_fragment (None)
        except KeyError:
            self.logger.error ('frame without url',
                    uuid='edc2743d-b93e-4ba1-964e-db232f2f96ff', tree=tree)
            url = None

        width = min (contentSize['width'], self.maxDim)
        # we’re ignoring horizontal scroll intentionally. Most horizontal
        # layouts use JavaScript scrolling and don’t extend the viewport.
        for yoff in range (0, contentHeight, self.maxDim):
            height = min (contentHeight - yoff, self.maxDim)
            clip = {'x': 0, 'y': yoff, 'width': width, 'height': height, 'scale': 1}
            ret = await tab.Page.captureScreenshot (format='png', clip=clip)
            data = b64decode (ret['data'])
            yield ScreenshotEvent (url, yoff, data)

        await tab.Emulation.clearDeviceMetricsOverride ()

class Click (JsOnload):
    """ Generic link clicking """

    name = 'click'
    scriptPath = 'click.js'

    def __init__ (self, loader, logger):
        super ().__init__ (loader, logger)
        with pkg_resources.resource_stream (__name__, os.path.join ('data', 'click.yaml')) as fd:
            self.options['sites'] = list (yaml.safe_load_all (fd))

class ExtractLinksEvent:
    __slots__ = ('links', )

    def __init__ (self, links):
        self.links = links

    def __repr__ (self):
        return f'<ExtractLinksEvent {self.links!r}>'

def mapOrIgnore (f, l):
    for e in l:
        try:
            yield f (e)
        except:
            pass

class ExtractLinks (Behavior):
    """
    Extract links from a page using JavaScript
    
    We could retrieve a HTML snapshot and extract links here, but we’d have to
    manually resolve relative links.
    """

    __slots__ = ('script', )

    name = 'extractLinks'

    def __init__ (self, loader, logger):
        super ().__init__ (loader, logger)
        self.script = Script ('extract-links.js')

    async def onfinish (self):
        tab = self.loader.tab
        yield self.script
        result = await tab.Runtime.evaluate (expression=str (self.script), returnByValue=True)
        yield ExtractLinksEvent (list (set (mapOrIgnore (URL, result['result']['value']))))

class Crash (Behavior):
    """ Crash the browser. For testing only. Obviously. """

    name = 'crash'

    async def onstop (self):
        try:
            await self.loader.tab.Page.crash ()
        except Crashed:
            pass
        return
        yield # pragma: no cover

# available behavior scripts. Order matters, move those modifying the page
# towards the end of available
available = [Scroll, Click, ExtractLinks, Screenshot, EmulateScreenMetrics, DomSnapshot]
#available.append (Crash)
# order matters, since behavior can modify the page (dom snapshots, for instance)
availableMap = OrderedDict (map (lambda x: (x.name, x), available))




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio
import pytest, html5lib
from html5lib.serializer import HTMLSerializer
from html5lib.treewalkers import getTreeWalker
from aiohttp import web

from .html import StripTagFilter, StripAttributeFilter, ChromeTreeWalker
from .test_devtools import tab, browser

def test_strip_tag ():
    d = html5lib.parse ('<a>barbaz<b>foobar</b>.</a><b>foobar</b>.<b attr=1><c></c>')
    stream = StripTagFilter (getTreeWalker ('etree')(d), ['b', 'c'])
    serializer = HTMLSerializer ()
    assert serializer.render (stream) == '<a>barbaz.</a>.'

def test_strip_attribute ():
    d = html5lib.parse ('<a b=1 c="yes" d></a><br b=2 c="no" d keep=1>')
    stream = StripAttributeFilter (getTreeWalker ('etree')(d), ['b', 'c', 'd'])
    serializer = HTMLSerializer ()
    assert serializer.render (stream) == '<a></a><br keep=1>'

@pytest.mark.asyncio
async def test_treewalker (tab):
    frames = await tab.Page.getFrameTree ()

    framehtml = '<HTML><HEAD></HEAD><BODY></BODY></HTML>'
    html = '<HTML><HEAD><META charset=utf-8></HEAD><BODY><H1>Hello</H1><!-- comment --><IFRAME></IFRAME></BODY></HTML>'
    rootframe = frames['frameTree']['frame']['id']
    await tab.Page.setDocumentContent (frameId=rootframe, html=html)

    dom = await tab.DOM.getDocument (depth=-1, pierce=True)
    docs = list (ChromeTreeWalker (dom['root']).split ())
    assert len(docs) == 2
    for i, doc in enumerate (docs):
        walker = ChromeTreeWalker (doc)
        serializer = HTMLSerializer ()
        result = serializer.render (iter(walker))
        if i == 0:
            assert result == html
        elif i == 1:
            assert result == framehtml

cdataDoc = '<test><![CDATA[Hello world]]></test>'
xmlHeader = '<?xml version="1.0" encoding="UTF-8"?>'
async def hello(request):
    return web.Response(text=xmlHeader + cdataDoc, content_type='text/xml')

@pytest.fixture
async def server ():
    """ Simple HTTP server for testing notifications """
    app = web.Application()
    app.add_routes([web.get('/test.xml', hello)])
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, 'localhost', 8080)
    await site.start()
    yield app
    await runner.cleanup ()

@pytest.mark.asyncio
async def test_treewalker_cdata (tab, server):
    ret = await tab.Page.navigate (url='http://localhost:8080/test.xml')
    # wait until loaded XXX: replace with idle check
    await asyncio.sleep (0.5)
    dom = await tab.DOM.getDocument (depth=-1, pierce=True)
    docs = list (ChromeTreeWalker (dom['root']).split ())
    assert len(docs) == 1
    for i, doc in enumerate (docs):
        walker = ChromeTreeWalker (doc)
        serializer = HTMLSerializer ()
        result = serializer.render (iter(walker))
        # chrome will display a pretty-printed viewer *plus* the original
        # source (stripped of its xml header)
        assert cdataDoc in result





# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio, socket
from operator import itemgetter
from http.server import BaseHTTPRequestHandler
from datetime import datetime

from yarl import URL
from aiohttp import web
from multidict import CIMultiDict

from hypothesis import given
import hypothesis.strategies as st
from hypothesis.provisional import domains
import pytest

from .browser import RequestResponsePair, SiteLoader, Request, \
        UnicodeBody, ReferenceTimestamp, Base64Body, UnicodeBody, Request, \
        Response, NavigateError, PageIdle, FrameNavigated
from .logger import Logger, Consumer
from .devtools import Crashed, Process

# if you want to know what’s going on:
#import logging
#logging.basicConfig(level=logging.DEBUG)

class AssertConsumer (Consumer):
    def __call__ (self, **kwargs):
        assert 'uuid' in kwargs
        assert 'msg' in kwargs
        assert 'context' in kwargs
        return kwargs

@pytest.fixture
def logger ():
    return Logger (consumer=[AssertConsumer ()])

@pytest.fixture
async def loader (logger):
    async with Process () as browser, SiteLoader (browser, logger) as l:
        yield l

@pytest.mark.asyncio
async def test_crash (loader):
    with pytest.raises (Crashed):
        await loader.tab.Page.crash ()

@pytest.mark.asyncio
async def test_invalidurl (loader):
    host = 'nonexistent.example'

    # make sure the url does *not* resolve (some DNS intercepting ISP’s mess
    # with this)
    loop = asyncio.get_event_loop ()
    try:
        resolved = await loop.getaddrinfo (host, None)
    except socket.gaierror:
        url = URL.build (scheme='http', host=host)
        with pytest.raises (NavigateError):
            await loader.navigate (url)
    else:
        pytest.skip (f'host {host} resolved to {resolved}')

timestamp = st.one_of (
                st.integers(min_value=0, max_value=2**32-1),
                st.floats (min_value=0, max_value=2**32-1),
                )

@given(timestamp, timestamp, timestamp)
def test_referencetimestamp (relativeA, absoluteA, relativeB):
    ts = ReferenceTimestamp (relativeA, absoluteA)
    absoluteA = datetime.utcfromtimestamp (absoluteA)
    absoluteB = ts (relativeB)
    assert (absoluteA < absoluteB and relativeA < relativeB) or \
            (absoluteA >= absoluteB and relativeA >= relativeB)
    assert abs ((absoluteB - absoluteA).total_seconds () - (relativeB - relativeA)) < 10e-6

def urls ():
    """ Build http/https URL """
    scheme = st.sampled_from (['http', 'https'])
    # Path must start with a slash
    pathSt = st.builds (lambda x: '/' + x, st.text ())
    args = st.fixed_dictionaries ({
            'scheme': scheme,
            'host': domains (),
            'port': st.one_of (st.none (), st.integers (min_value=1, max_value=2**16-1)),
            'path': pathSt,
            'query_string': st.text (),
            'fragment': st.text (),
            })
    return st.builds (lambda x: URL.build (**x), args)

def urlsStr ():
    return st.builds (lambda x: str (x), urls ())

asciiText = st.text (st.characters (min_codepoint=32, max_codepoint=126))

def chromeHeaders ():
    # token as defined by https://tools.ietf.org/html/rfc7230#section-3.2.6
    token = st.sampled_from('abcdefghijklmnopqrstuvwxyz0123456789!#$%&\'*+-.^_`|~')
    # XXX: the value should be asciiText without leading/trailing spaces
    return st.dictionaries (token, token)

def fixedDicts (fixed, dynamic):
    return st.builds (lambda x, y: x.update (y), st.fixed_dictionaries (fixed), st.lists (dynamic))

def chromeRequestWillBeSent (reqid, url):
    methodSt = st.sampled_from (['GET', 'POST', 'PUT', 'DELETE'])
    return st.fixed_dictionaries ({
            'requestId': reqid,
            'initiator': st.just ('Test'),
            'wallTime': timestamp,
            'timestamp': timestamp,
            'request': st.fixed_dictionaries ({
                'url': url,
                'method': methodSt,
                'headers': chromeHeaders (),
                # XXX: postData, hasPostData
                })
            })

def chromeResponseReceived (reqid, url):
    mimeTypeSt = st.one_of (st.none (), st.just ('text/html'))
    remoteIpAddressSt = st.one_of (st.none (), st.just ('127.0.0.1'))
    protocolSt = st.one_of (st.none (), st.just ('h2'))
    statusCodeSt = st.integers (min_value=100, max_value=999)
    typeSt = st.sampled_from (['Document', 'Stylesheet', 'Image', 'Media',
            'Font', 'Script', 'TextTrack', 'XHR', 'Fetch', 'EventSource',
            'WebSocket', 'Manifest', 'SignedExchange', 'Ping',
            'CSPViolationReport', 'Other'])
    return st.fixed_dictionaries ({
            'requestId': reqid,
            'timestamp': timestamp,
            'type': typeSt,
            'response': st.fixed_dictionaries ({
                'url': url,
                'requestHeaders': chromeHeaders (), # XXX: make this optional
                'headers': chromeHeaders (),
                'status': statusCodeSt,
                'statusText': asciiText,
                'mimeType': mimeTypeSt,
                'remoteIPAddress': remoteIpAddressSt,
                'protocol': protocolSt,
                })
            })

def chromeReqResp ():
    # XXX: will this gnerated the same url for all testcases?
    reqid = st.shared (st.text (), 'reqresp')
    url = st.shared (urlsStr (), 'reqresp')
    return st.tuples (chromeRequestWillBeSent (reqid, url),
            chromeResponseReceived (reqid, url))

def requestResponsePair ():
    def f (creq, cresp, hasPostData, reqBody, respBody):
        i = RequestResponsePair ()
        i.fromRequestWillBeSent (creq)
        i.request.hasPostData = hasPostData
        if hasPostData:
            i.request.body = reqBody

        if cresp is not None:
            i.fromResponseReceived (cresp)
            if respBody is not None:
                i.response.body = respBody
        return i

    bodySt = st.one_of (
            st.none (),
            st.builds (UnicodeBody, st.text ()),
            st.builds (Base64Body.fromBytes, st.binary ())
            )
    return st.builds (lambda reqresp, hasPostData, reqBody, respBody:
            f (reqresp[0], reqresp[1], hasPostData, reqBody, respBody),
            chromeReqResp (), st.booleans (), bodySt, bodySt)

@given(chromeReqResp ())
def test_requestResponsePair (creqresp):
    creq, cresp = creqresp

    item = RequestResponsePair ()

    assert item.id is None
    assert item.url is None
    assert item.request is None
    assert item.response is None

    item.fromRequestWillBeSent (creq)

    assert item.id == creq['requestId']
    url = URL (creq['request']['url'])
    assert item.url == url
    assert item.request is not None
    assert item.request.timestamp == datetime.utcfromtimestamp (creq['wallTime'])
    assert set (item.request.headers.keys ()) == set (creq['request']['headers'].keys ())
    assert item.response is None

    item.fromResponseReceived (cresp)

    # url will not be overwritten
    assert item.id == creq['requestId'] == cresp['requestId']
    assert item.url == url
    assert item.request is not None
    assert set (item.request.headers.keys ()) == set (cresp['response']['requestHeaders'].keys ())
    assert item.response is not None
    assert set (item.response.headers.keys ()) == set (cresp['response']['headers'].keys ())
    assert (item.response.timestamp - item.request.timestamp).total_seconds () - \
            (cresp['timestamp'] - creq['timestamp']) < 10e-6

@given(chromeReqResp ())
def test_requestResponsePair_eq (creqresp):
    creq, cresp = creqresp

    item = RequestResponsePair ()
    item2 = RequestResponsePair ()
    assert item == item
    assert item == item2

    item.fromRequestWillBeSent (creq)
    assert item != item2
    item2.fromRequestWillBeSent (creq)
    assert item == item
    assert item == item2

    item.fromResponseReceived (cresp)
    assert item != item2
    item2.fromResponseReceived (cresp)
    assert item == item
    assert item == item2

    # XXX: test for inequality with different parameters

### Google Chrome integration tests ###

serverUrl = URL.build (scheme='http', host='localhost', port=8080)
items = [
    RequestResponsePair (
        url=serverUrl.with_path ('/encoding/utf-8'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/html; charset=utf-8')]),
            body=UnicodeBody ('äöü'), mimeType='text/html')
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/encoding/latin1'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/html; charset=latin1')]),
            body=UnicodeBody ('äöü'), mimeType='text/html')
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/encoding/utf-16'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/html; charset=utf-16')]),
            body=UnicodeBody ('äöü'), mimeType='text/html')
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/encoding/ISO-8859-1'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/html; charset=ISO-8859-1')]),
            body=UnicodeBody ('äöü'), mimeType='text/html')
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/status/200'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/plain')]),
            body=b'',
            mimeType='text/plain'),
        ),
    # redirects never have a response body
    RequestResponsePair (
        url=serverUrl.with_path ('/status/301'),
        request=Request (method='GET'),
        response=Response (status=301,
            headers=CIMultiDict ([('Content-Type', 'text/plain'),
                ('Location', str (serverUrl.with_path ('/status/301/redirected')))]),
            body=None,
            mimeType='text/plain'),
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/image/png'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'image/png')]),
            body=Base64Body.fromBytes (b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x01\x00\x00\x00\x01\x08\x00\x00\x00\x00:~\x9bU\x00\x00\x00\nIDAT\x08\x1dc\xf8\x0f\x00\x01\x01\x01\x006_g\x80\x00\x00\x00\x00IEND\xaeB`\x82'),
            mimeType='image/png'),
        ),
    RequestResponsePair (
        url=serverUrl.with_path ('/script/alert'),
        request=Request (method='GET'),
        response=Response (status=200, headers=CIMultiDict ([('Content-Type', 'text/html; charset=utf-8')]),
            body=UnicodeBody ('''<html><body><script>
window.addEventListener("beforeunload", function (e) {
    e.returnValue = "bye?";
    return e.returnValue;
});
alert("stopping here");
if (confirm("are you sure?") || prompt ("42?")) {
    window.location = "/nonexistent";
}
</script></body></html>'''), mimeType='text/html')
        ),
    ]

@pytest.mark.asyncio
# would be nice if we could use hypothesis here somehow
@pytest.mark.parametrize("golden", items)
async def test_integration_item (loader, golden):
    async def f (req):
        body = golden.response.body
        contentType = golden.response.headers.get ('content-type', '') if golden.response.headers is not None else ''
        charsetOff = contentType.find ('charset=')
        if isinstance (body, UnicodeBody) and charsetOff != -1:
            encoding = contentType[charsetOff+len ('charset='):]
            body = golden.response.body.decode ('utf-8').encode (encoding)
        return web.Response (body=body, status=golden.response.status,
                headers=golden.response.headers)

    app = web.Application ()
    app.router.add_route (golden.request.method, golden.url.path, f)
    runner = web.AppRunner(app)
    await runner.setup()
    site = web.TCPSite(runner, serverUrl.host, serverUrl.port)
    try:
        await site.start()
    except Exception as e:
        pytest.skip (e)

    haveReqResp = False
    haveNavigated = False
    try:
        await loader.navigate (golden.url)

        it = loader.__aiter__ ()
        while True:
            try:
                item = await asyncio.wait_for (it.__anext__ (), timeout=1)
            except asyncio.TimeoutError:
                break
            # XXX: can only check the first req/resp right now (due to redirect)
            if isinstance (item, RequestResponsePair) and not haveReqResp:
                # we do not know this in advance
                item.request.initiator = None
                item.request.headers = None
                item.remoteIpAddress = None
                item.protocol = None
                item.resourceType = None

                if item.response:
                    assert item.response.statusText is not None
                    item.response.statusText = None

                    del item.response.headers['server']
                    del item.response.headers['content-length']
                    del item.response.headers['date']
                assert item == golden
                haveReqResp = True
            elif isinstance (item, FrameNavigated):
                # XXX: can’t check this, because of the redirect
                #assert item.url == golden.url
                haveNavigated = True
    finally:
        assert haveReqResp
        assert haveNavigated
        await runner.cleanup ()

def test_page_idle ():
    for v in (True, False):
        idle = PageIdle (v)
        assert bool (idle) == v





# Copyright (c) 2017–2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Simple logger inspired by structlog.

It is usually used like this: Classes are passed a logger instance. They bind
context to their name, so identifying the source of messages is easier. Every
log message carries a unique id (uuid) for automated identification as well as
a short human-readable message (msg) and arbitrary payload.
"""

import sys, json
from datetime import datetime
from functools import partial
from enum import IntEnum

from pytz import utc

from .util import StrJsonEncoder

class Level(IntEnum):
    DEBUG = 0
    INFO = 1
    WARNING = 2
    ERROR = 3

class Logger:
    def __init__ (self, consumer=None, bindings=None):
        self.bindings = bindings or {}
        self.consumer = consumer or []

    def __call__ (self, level, *args, **kwargs):
        if not isinstance (level, Level):
            level = Level[level.upper ()]
        kwargs['level'] = level
        if args:
            if len (args) == 1:
                args, = args
            kwargs['msg'] = args
        # do not overwrite arguments
        for k, v in self.bindings.items ():
            if k not in kwargs:
                kwargs[k] = v
        for c in self.consumer:
            kwargs = c (**kwargs)
        return kwargs

    def __getattr__ (self, k):
        """ Bind all method names to level, so Logger.info, Logger.warning, … work """
        return partial (self.__call__, k)

    def bind (self, **kwargs):
        d = self.bindings.copy ()
        d.update (kwargs)
        # consumer is not a copy intentionally, so attaching to the parent
        # logger will attach to all children as well
        return self.__class__ (consumer=self.consumer, bindings=d)

    def unbind (self, **kwargs):
        d = self.bindings.copy ()
        for k in kwargs.keys ():
            del d[k]
        return self.__class__ (consumer=self.consumer, bindings=d)

    def connect (self, consumer):
        self.consumer.append (consumer)

    def disconnect (self, consumer):
        self.consumer.remove (consumer)

class Consumer:
    def __call__ (self, **kwargs): # pragma: no cover
        raise NotImplementedError ()

class NullConsumer (Consumer):
    def __call__ (self, **kwargs):
        return kwargs

class PrintConsumer (Consumer):
    """
    Simple printing consumer
    """
    def __call__ (self, **kwargs):
        sys.stderr.write (str (kwargs))
        sys.stderr.write ('\n')
        sys.stderr.flush ()
        return kwargs

class JsonPrintConsumer (Consumer):
    def __init__ (self, minLevel=Level.DEBUG):
        self.minLevel = minLevel

    def __call__ (self, **kwargs):
        if kwargs['level'] >= self.minLevel:
            json.dump (kwargs, sys.stdout, cls=StrJsonEncoder)
            sys.stdout.write ('\n')
            sys.stdout.flush ()
        return kwargs

class DatetimeConsumer (Consumer):
    def __call__ (self, **kwargs):
        kwargs['date'] = datetime.utcnow ().replace (tzinfo=utc)
        return kwargs

class WarcHandlerConsumer (Consumer):
    def __init__ (self, warc, minLevel=Level.DEBUG):
        self.warc = warc
        self.minLevel = minLevel

    def __call__ (self, **kwargs):
        if kwargs['level'] >= self.minLevel:
            self.warc._writeLog (json.dumps (kwargs, cls=StrJsonEncoder))
        return kwargs




# Copyright (c) 2018 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from tempfile import NamedTemporaryFile
from operator import itemgetter
from io import BytesIO
import pytest
from warcio.archiveiterator import ArchiveIterator
from warcio.warcwriter import WARCWriter
from warcio.statusandheaders import StatusAndHeaders
from pkg_resources import parse_version

from .tools import mergeWarc, Errata, FixableErrata

@pytest.fixture
def writer():
    return WARCWriter (NamedTemporaryFile(), gzip=True)

def recordsEqual(golden, underTest):
    for a, b in zip (golden, underTest):
        # record ids are not predictable, so we cannot compare them. Dito for
        # dates. Content-* seems to be added when writing to file.
        for x in {'WARC-Record-Id', 'WARC-Block-Digest', 'WARC-Date',
                'Content-Length', 'Content-Type'}:
            a.rec_headers.remove_header(x)
            b.rec_headers.remove_header(x)
        aheader = sorted(a.rec_headers.headers, key=itemgetter(0))
        bheader = sorted(b.rec_headers.headers, key=itemgetter(0))
        assert aheader == bheader
        assert a.http_headers == b.http_headers

def makeGolden(writer, records):
    # additional warcinfo is written. Content does not matter.
    record = writer.create_warc_record (
            '',
            'warcinfo',
            payload=b'',
            warc_headers_dict={'Content-Type': 'text/plain; encoding=utf-8'})
    records.insert (0, record)
    return records

def test_unmodified(writer):
    """
    Single request/response pair, no revisits
    """

    records = []

    httpHeaders = StatusAndHeaders('GET / HTTP/1.1', {}, is_http_request=True)
    warcHeaders = {}
    record = writer.create_warc_record ('http://example.com/', 'request', payload=BytesIO(b'foobar'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    httpHeaders = StatusAndHeaders('200 OK', {}, protocol='HTTP/1.1')
    record = writer.create_warc_record ('http://example.com/', 'response', payload=BytesIO(b'data'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    for r in records:
        writer.write_record (r)

    output = NamedTemporaryFile()
    mergeWarc ([writer.out.name], output)

    output.seek(0)
    recordsEqual (makeGolden (writer, records), ArchiveIterator (output))

def test_different_payload(writer):
    """
    Duplicate URL, but different payload
    """

    records = []
    for i in range (2):
        httpHeaders = StatusAndHeaders('GET / HTTP/1.1', {}, is_http_request=True)
        warcHeaders = {}
        record = writer.create_warc_record ('http://example.com/', 'request', payload=BytesIO(b'foobar'),
                warc_headers_dict=warcHeaders, http_headers=httpHeaders)
        records.append (record)

        httpHeaders = StatusAndHeaders('200 OK', {}, protocol='HTTP/1.1')
        record = writer.create_warc_record ('http://example.com/', 'response',
                payload=BytesIO(f'data{i}'.encode ('utf8')),
                warc_headers_dict=warcHeaders, http_headers=httpHeaders)
        records.append (record)

    for r in records:
        writer.write_record (r)

    output = NamedTemporaryFile()
    mergeWarc ([writer.out.name], output)

    output.seek(0)
    recordsEqual (makeGolden (writer, records), ArchiveIterator (output))

def makeRevisit(writer, ref, dup):
    """ Make revisit record for reference """
    dupHeaders = dup.rec_headers
    refHeaders = ref.rec_headers
    record = writer.create_revisit_record (dupHeaders.get_header('WARC-Target-URI'),
            digest=refHeaders.get_header('WARC-Payload-Digest'),
            refers_to_uri=refHeaders.get_header('WARC-Target-URI'),
            refers_to_date=refHeaders.get_header('WARC-Date'),
            http_headers=dup.http_headers)
    record.rec_headers.add_header ('WARC-Refers-To', refHeaders.get_header('WARC-Record-ID'))
    record.rec_headers.add_header ('WARC-Truncated', 'length')
    return record

def test_resp_revisit_same_url(writer):
    """
    Duplicate record for the same URL, creates a revisit
    """

    records = []
    for i in range (2):
        httpHeaders = StatusAndHeaders('GET / HTTP/1.1', {}, is_http_request=True)
        warcHeaders = {}
        record = writer.create_warc_record ('http://example.com/', 'request', payload=BytesIO(b'foobar'),
                warc_headers_dict=warcHeaders, http_headers=httpHeaders)
        records.append (record)

        httpHeaders = StatusAndHeaders('200 OK', {}, protocol='HTTP/1.1')
        record = writer.create_warc_record ('http://example.com/', 'response', payload=BytesIO(b'data'),
                warc_headers_dict=warcHeaders, http_headers=httpHeaders)
        records.append (record)

    for r in records:
        writer.write_record (r)

    dup = records.pop ()
    ref = records[1]
    records.append (makeRevisit (writer, ref, dup))

    output = NamedTemporaryFile()
    mergeWarc ([writer.out.name], output)

    output.seek(0)
    recordsEqual (makeGolden (writer, records), ArchiveIterator (output))

def test_resp_revisit_other_url(writer):
    """
    Duplicate record for different URL, creates a revisit
    """

    records = []

    httpHeaders = StatusAndHeaders('GET / HTTP/1.1', {}, is_http_request=True)
    warcHeaders = {}
    record = writer.create_warc_record ('http://example.com/', 'request', payload=BytesIO(b'foobar'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    httpHeaders = StatusAndHeaders('200 OK', {}, protocol='HTTP/1.1')
    record = writer.create_warc_record ('http://example.com/', 'response', payload=BytesIO(b'data'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    httpHeaders = StatusAndHeaders('GET / HTTP/1.1', {}, is_http_request=True)
    warcHeaders = {}
    record = writer.create_warc_record ('http://example.com/one', 'request', payload=BytesIO(b'foobar'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    httpHeaders = StatusAndHeaders('200 OK', {}, protocol='HTTP/1.1')
    record = writer.create_warc_record ('http://example.com/one', 'response', payload=BytesIO(b'data'),
            warc_headers_dict=warcHeaders, http_headers=httpHeaders)
    records.append (record)

    for r in records:
        writer.write_record (r)

    dup = records.pop ()
    ref = records[1]
    records.append (makeRevisit (writer, ref, dup))

    output = NamedTemporaryFile()
    mergeWarc ([writer.out.name], output)

    output.seek(0)
    recordsEqual (makeGolden (writer, records), ArchiveIterator (output))

def test_errata_contains():
    """ Test version matching """
    e = Errata('some-uuid', 'description', ['a<1.0'])
    assert {'a': parse_version('0.1')} in e
    assert {'a': parse_version('1.0')} not in e
    assert {'b': parse_version('1.0')} not in e

    e = Errata('some-uuid', 'description', ['a<1.0,>0.1'])
    assert {'a': parse_version('0.1')} not in e
    assert {'a': parse_version('0.2')} in e
    assert {'a': parse_version('1.0')} not in e

    # a AND b
    e = Errata('some-uuid', 'description', ['a<1.0', 'b>1.0'])
    assert {'a': parse_version('0.1')} not in e
    assert {'b': parse_version('1.1')} not in e
    assert {'a': parse_version('0.1'), 'b': parse_version('1.1')} in e

def test_errata_fixable ():
    e = Errata('some-uuid', 'description', ['a<1.0', 'b>1.0'])
    assert not e.fixable

    e = FixableErrata('some-uuid', 'description', ['a<1.0', 'b>1.0'])
    assert e.fixable




# Copyright (c) 2017 crocoite contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Classes writing data to WARC files
"""

import json, threading
from io import BytesIO
from datetime import datetime
from http.server import BaseHTTPRequestHandler

from warcio.timeutils import datetime_to_iso_date
from warcio.warcwriter import WARCWriter
from warcio.statusandheaders import StatusAndHeaders
from yarl import URL

from .util import StrJsonEncoder
from .controller import EventHandler, ControllerStart
from .behavior import Script, DomSnapshotEvent, ScreenshotEvent
from .browser import RequestResponsePair, UnicodeBody

# the official mimetype for json, according to https://tools.ietf.org/html/rfc8259
jsonMime = 'application/json'
# mime for javascript, according to https://tools.ietf.org/html/rfc4329#section-7.2
jsMime = 'application/javascript'

class WarcHandler (EventHandler):
    __slots__ = ('logger', 'writer', 'documentRecords', 'log',
            'maxLogSize', 'logEncoding', 'warcinfoRecordId')

    def __init__ (self, fd, logger):
        self.logger = logger
        self.writer = WARCWriter (fd, gzip=True)

        self.logEncoding = 'utf-8'
        self.log = BytesIO ()
        # max log buffer size (bytes)
        self.maxLogSize = 500*1024

        # maps document urls to WARC record ids, required for DomSnapshotEvent
        # and ScreenshotEvent
        self.documentRecords = {}
        # record id of warcinfo record
        self.warcinfoRecordId = None

    def __enter__ (self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._flushLogEntries ()

    def writeRecord (self, url, kind, payload, warc_headers_dict=None, http_headers=None):
        """
        Thin wrapper around writer.create_warc_record and writer.write_record.

        Adds default WARC headers.
        """
        assert url is None or isinstance (url, URL)

        d = {}
        if self.warcinfoRecordId:
            d['WARC-Warcinfo-ID'] = self.warcinfoRecordId
        d.update (warc_headers_dict)
        warc_headers_dict = d

        record = self.writer.create_warc_record (str (url) if url else '',
                kind,
                payload=payload,
                warc_headers_dict=warc_headers_dict,
                http_headers=http_headers)
        self.writer.write_record (record)

        return record

    def _writeRequest (self, item):
        logger = self.logger.bind (reqId=item.id)

        req = item.request
        url = item.url

        path = url.relative().with_fragment(None)
        httpHeaders = StatusAndHeaders(f'{req.method} {path} HTTP/1.1',
                req.headers, protocol='HTTP/1.1', is_http_request=True)
        warcHeaders = {
                # required to correlate request with log entries
                'X-Chrome-Request-ID': item.id,
                'WARC-Date': datetime_to_iso_date (req.timestamp),
                }

        body = item.request.body
        if item.request.hasPostData and body is None:
            # oops, don’t know what went wrong here
            logger.error ('requestBody missing',
                    uuid='ee9adc58-e723-4595-9feb-312a67ead6a0')
            warcHeaders['WARC-Truncated'] = 'unspecified'
        else:
            body = BytesIO (body)
        record = self.writeRecord (url, 'request',
                payload=body, http_headers=httpHeaders,
                warc_headers_dict=warcHeaders)
        return record.rec_headers['WARC-Record-ID']

    def _writeResponse (self, item, concurrentTo):
        # fetch the body
        reqId = item.id

        # now the response
        resp = item.response
        warcHeaders = {
                'WARC-Concurrent-To': concurrentTo,
                # required to correlate request with log entries
                'X-Chrome-Request-ID': item.id,
                'WARC-Date': datetime_to_iso_date (resp.timestamp),
                }
        # conditional WARC headers
        if item.remoteIpAddress:
            warcHeaders['WARC-IP-Address'] = item.remoteIpAddress

        # HTTP headers
        statusText = resp.statusText or \
                BaseHTTPRequestHandler.responses.get (
                resp.status, ('No status text available', ))[0]
        httpHeaders = StatusAndHeaders(f'{resp.status} {statusText}',
                resp.headers, protocol='HTTP/1.1')

        # Content is saved decompressed and decoded, remove these headers
        blacklistedHeaders = {'transfer-encoding', 'content-encoding'}
        for h in blacklistedHeaders:
            httpHeaders.remove_header (h)

        # chrome sends nothing but utf8 encoded text. Fortunately HTTP
        # headers take precedence over the document’s <meta>, thus we can
        # easily override those.
        contentType = resp.mimeType
        if contentType:
            if isinstance (resp.body, UnicodeBody):
                contentType += '; charset=utf-8'
            httpHeaders.replace_header ('Content-Type', contentType)

        # response body
        body = resp.body
        if body is None:
            warcHeaders['WARC-Truncated'] = 'unspecified'
        else:
            httpHeaders.replace_header ('Content-Length', str (len (body)))
            body = BytesIO (body)

        record = self.writeRecord (item.url, 'response',
                warc_headers_dict=warcHeaders, payload=body,
                http_headers=httpHeaders)

        if item.resourceType == 'Document':
            self.documentRecords[item.url] = record.rec_headers.get_header ('WARC-Record-ID')

    def _writeScript (self, item):
        writer = self.writer
        encoding = 'utf-8'
        # XXX: yes, we’re leaking information about the user here, but this is
        # the one and only source URL of the scripts.
        uri = URL(f'file://{item.abspath}') if item.path else None
        self.writeRecord (uri, 'resource',
                payload=BytesIO (str (item).encode (encoding)),
                warc_headers_dict={
                    'Content-Type': f'{jsMime}; charset={encoding}',
                    'X-Crocoite-Type': 'script',
                    })

    def _writeItem (self, item):
        assert item.request
        concurrentTo = self._writeRequest (item)
        # items that failed loading don’t have a response
        if item.response:
            self._writeResponse (item, concurrentTo)

    def _addRefersTo (self, headers, url):
        refersTo = self.documentRecords.get (url)
        if refersTo:
            headers['WARC-Refers-To'] = refersTo
        else:
            self.logger.error (f'No document record found for {url}')
        return headers

    def _writeDomSnapshot (self, item):
        writer = self.writer

        warcHeaders = {
                'X-Crocoite-Type': 'dom-snapshot',
                'X-Chrome-Viewport': item.viewport,
                'Content-Type': 'text/html; charset=utf-8',
                }

        self._addRefersTo (warcHeaders, item.url)

        self.writeRecord (item.url, 'conversion',
                payload=BytesIO (item.document),
                warc_headers_dict=warcHeaders)

    def _writeScreenshot (self, item):
        writer = self.writer
        warcHeaders = {
                'Content-Type': 'image/png',
                'X-Crocoite-Screenshot-Y-Offset': str (item.yoff),
                'X-Crocoite-Type': 'screenshot',
                }
        self._addRefersTo (warcHeaders, item.url)
        self.writeRecord (item.url, 'conversion',
                payload=BytesIO (item.data), warc_headers_dict=warcHeaders)

    def _writeControllerStart (self, item, encoding='utf-8'):
        payload = BytesIO (json.dumps (item.payload, indent=2, cls=StrJsonEncoder).encode (encoding))

        writer = self.writer
        warcinfo = self.writeRecord (None, 'warcinfo',
                warc_headers_dict={'Content-Type': f'{jsonMime}; encoding={encoding}'},
                payload=payload)
        self.warcinfoRecordId = warcinfo.rec_headers['WARC-Record-ID']

    def _flushLogEntries (self):
        if self.log.tell () > 0:
            writer = self.writer
            self.log.seek (0)
            warcHeaders = {
                    'Content-Type': f'application/json; encoding={self.logEncoding}',
                    'X-Crocoite-Type': 'log',
                    }
            self.writeRecord (None, 'metadata', payload=self.log,
                    warc_headers_dict=warcHeaders)
            self.log = BytesIO ()

    def _writeLog (self, item):
        """ Handle log entries, called by .logger.WarcHandlerConsumer only """
        self.log.write (item.encode (self.logEncoding))
        self.log.write (b'\n')
        if self.log.tell () > self.maxLogSize:
            self._flushLogEntries ()

    route = {Script: _writeScript,
            RequestResponsePair: _writeItem,
            DomSnapshotEvent: _writeDomSnapshot,
            ScreenshotEvent: _writeScreenshot,
            ControllerStart: _writeControllerStart,
            }

    async def push (self, item):
        for k, v in self.route.items ():
            if isinstance (item, k):
                v (self, item)
                break

